"""Plugin registry and chooser for managing ptychodus extensions."""

from __future__ import annotations
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Generic, TypeVar
import importlib
import logging
import pkgutil
import re

from .diffraction import BadPixelsFileReader, DiffractionFileReader, DiffractionFileWriter
from .fluorescence import (
    DeconvolutionStrategy,
    FluorescenceFileReader,
    FluorescenceFileWriter,
    UpscalingStrategy,
)
from .object import ObjectFileReader, ObjectFileWriter, Object
from .observer import Observable, Observer
from .parametric import StringParameter
from .probe import FresnelZonePlate, ProbeFileReader, ProbeFileWriter, ProbeSequence
from .product import ProductFileReader, ProductFileWriter
from .probe_positions import (
    ProbePositionFileReader,
    ProbePositionFileWriter,
    ProbePositionSequence,
)
from .workflow import FileBasedWorkflow

__all__ = [
    'PluginChooser',
    'PluginRegistry',
]

T = TypeVar('T')

logger = logging.getLogger(__name__)


class ProductProbePositionFileReader(ProbePositionFileReader):
    """Adapter that extracts the probe positions from a ProductFileReader."""

    def __init__(self, reader: ProductFileReader) -> None:
        super().__init__()
        self._reader = reader

    def read(self, file_path: Path) -> ProbePositionSequence:
        product = self._reader.read(file_path)
        return product.probe_positions


class ProductProbeFileReader(ProbeFileReader):
    """Adapter that extracts the probe sequence from a ProductFileReader."""

    def __init__(self, reader: ProductFileReader) -> None:
        super().__init__()
        self._reader = reader

    def read(self, file_path: Path) -> ProbeSequence:
        product = self._reader.read(file_path)
        return product.probes


class ProductObjectFileReader(ObjectFileReader):
    """Adapter that extracts the object from a ProductFileReader."""

    def __init__(self, reader: ProductFileReader) -> None:
        super().__init__()
        self._reader = reader

    def read(self, file_path: Path) -> Object:
        product = self._reader.read(file_path)
        return product.object_


@dataclass(frozen=True)
class Plugin(Generic[T]):
    """A registered plugin: its strategy object and both simple and display names."""

    strategy: T
    simple_name: str
    display_name: str


class PluginChooser(Iterable[Plugin[T]], Observable, Observer):
    """Observable list of typed plugins with a tracked current selection."""

    def __init__(self) -> None:
        super().__init__()
        self._registered_plugins: list[Plugin[T]] = list()
        self._current_index = 0
        self._parameter: StringParameter | None = None

    def stringify_plugin_names(self) -> str:
        return ', '.join(sorted(plugin.simple_name for plugin in self._registered_plugins))

    def register_plugin(self, strategy: T, *, display_name: str, simple_name: str = '') -> None:
        if not simple_name:
            simple_name = re.sub(r'\W+', '', display_name)

        plugin = Plugin[T](strategy, simple_name, display_name)
        self._registered_plugins.append(plugin)
        self._registered_plugins.sort(key=lambda x: x.display_name)
        self.notify_observers()

    def get_current_plugin(self) -> Plugin[T]:
        return self._registered_plugins[self._current_index]

    def set_current_plugin(self, name: str) -> None:
        namecf = name.casefold()

        for index, plugin in enumerate(self._registered_plugins):
            if namecf == plugin.simple_name.casefold() or namecf == plugin.display_name.casefold():
                if self._current_index != index:
                    self._current_index = index

                    if self._parameter is not None:
                        self._parameter.set_value(self.get_current_plugin().simple_name)

                    self.notify_observers()

                return

        registered_plugins = ', '.join(f'"{pi.simple_name}"' for pi in self._registered_plugins)
        logger.warning(f'Invalid plugin name "{name}". Registered plugins: {registered_plugins}.')

    def synchronize_with_parameter(self, parameter: StringParameter) -> None:
        self._parameter = parameter
        self.set_current_plugin(parameter.get_value())
        self._parameter.add_observer(self)

    def __iter__(self) -> Iterator[Plugin[T]]:
        for plugin in self._registered_plugins:
            yield plugin

    def __bool__(self) -> bool:
        return bool(self._registered_plugins)

    def _update(self, observable: Observable) -> None:
        if self._parameter is not None and observable is self._parameter:
            self.set_current_plugin(self._parameter.get_value())


class PluginRegistry:
    """Central collection of PluginChooser instances for every plugin category."""

    def __init__(self) -> None:
        self.bad_pixels_file_readers = PluginChooser[BadPixelsFileReader]()
        self.diffraction_file_readers = PluginChooser[DiffractionFileReader]()
        self.diffraction_file_writers = PluginChooser[DiffractionFileWriter]()
        self.probe_position_file_readers = PluginChooser[ProbePositionFileReader]()
        self.probe_position_file_writers = PluginChooser[ProbePositionFileWriter]()
        self.fresnel_zone_plates = PluginChooser[FresnelZonePlate]()
        self.probe_file_readers = PluginChooser[ProbeFileReader]()
        self.probe_file_writers = PluginChooser[ProbeFileWriter]()
        self.object_file_readers = PluginChooser[ObjectFileReader]()
        self.object_file_writers = PluginChooser[ObjectFileWriter]()
        self.product_file_readers = PluginChooser[ProductFileReader]()
        self.product_file_writers = PluginChooser[ProductFileWriter]()
        self.file_based_workflows = PluginChooser[FileBasedWorkflow]()
        self.fluorescence_file_readers = PluginChooser[FluorescenceFileReader]()
        self.fluorescence_file_writers = PluginChooser[FluorescenceFileWriter]()
        self.upscaling_strategies = PluginChooser[UpscalingStrategy]()
        self.deconvolution_strategies = PluginChooser[DeconvolutionStrategy]()

    def register_product_file_reader_with_adapters(
        self, strategy: ProductFileReader, *, display_name: str, simple_name: str = ''
    ) -> None:
        self.probe_position_file_readers.register_plugin(
            ProductProbePositionFileReader(strategy),
            display_name=display_name,
            simple_name=simple_name,
        )
        self.probe_file_readers.register_plugin(
            ProductProbeFileReader(strategy), display_name=display_name, simple_name=simple_name
        )
        self.object_file_readers.register_plugin(
            ProductObjectFileReader(strategy), display_name=display_name, simple_name=simple_name
        )
        self.product_file_readers.register_plugin(
            strategy, display_name=display_name, simple_name=simple_name
        )

    @classmethod
    def load_plugins(cls) -> PluginRegistry:
        registry = cls()

        import ptychodus.plugins

        ns_pkg: ModuleType = ptychodus.plugins

        # Specifying the second argument (prefix) to iter_modules makes the
        # returned name an absolute name instead of a relative one. This allows
        # import_module to work without having to do additional modification to
        # the name.
        for module_info in pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + '.'):
            try:
                module = importlib.import_module(module_info.name)
            except ModuleNotFoundError as exc:
                logger.warning(exc)
                logger.warning(f'Skipping {module_info.name}')
            else:
                try:
                    module.register_plugins(registry)
                except AttributeError as exc:
                    logger.warning(exc)
                    logger.warning(f'Failed to register {module_info.name}')
                else:
                    logger.debug(f'Registered {module_info.name}')

        return registry



from __future__ import annotations
from enum import auto, Enum
import logging

import numpy

from PyQt5.QtCore import pyqtSignal, Qt, QObject, QPointF, QLineF, QRectF, QSize, QSizeF
from PyQt5.QtGui import QIcon, QImage, QPalette, QPen, QPixmap, QWheelEvent
from PyQt5.QtWidgets import (
    QAction,
    QApplication,
    QComboBox,
    QDialog,
    QFormLayout,
    QGraphicsLineItem,
    QGraphicsPixmapItem,
    QGraphicsRectItem,
    QGraphicsSceneHoverEvent,
    QGraphicsSceneMouseEvent,
    QGraphicsView,
    QGroupBox,
    QLineEdit,
    QStatusBar,
    QToolBar,
    QVBoxLayout,
    QWidget,
)

from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure

from ptychodus.api.visualization import VisualizationProduct

from .widgets import DecimalLineEdit

logger = logging.getLogger(__name__)


class ImageItemSignals(QObject):
    rectangle_finished = pyqtSignal(QRectF)
    line_cut_finished = pyqtSignal(QLineF)
    fourier_finished = pyqtSignal(QRectF)


class ImageMouseTool(Enum):
    MOVE_TOOL = auto()
    RULER_TOOL = auto()
    RECTANGLE_TOOL = auto()
    LINE_CUT_TOOL = auto()
    FOURIER_TOOL = auto()


class ImageItem(QGraphicsPixmapItem):
    def __init__(self, signals: ImageItemSignals, status_bar: QStatusBar) -> None:
        super().__init__()
        self._signals = signals
        self._status_bar = status_bar
        self._product: VisualizationProduct | None = None
        self._mouse_tool = ImageMouseTool.MOVE_TOOL

        pen = QPen(Qt.PenStyle.DashLine)
        pen.setCapStyle(Qt.PenCapStyle.FlatCap)
        pen.setJoinStyle(Qt.PenJoinStyle.MiterJoin)
        pen.setCosmetic(True)

        self._line_item = QGraphicsLineItem(self)
        self._line_item.setPen(pen)
        self._line_item.hide()
        self._line_item.setZValue(100)

        self._rectangle_item = QGraphicsRectItem(self)
        self._rectangle_item.setPen(pen)
        self._rectangle_item.hide()
        self._rectangle_item.setZValue(100)
        self._rectangle_origin = QPointF()

        self._fourier_item = QGraphicsRectItem(self)
        self._fourier_item.setPen(pen)
        self._fourier_item.hide()
        self._fourier_item.setZValue(100)
        self._fourier_origin = QPointF()

        self.setTransformationMode(Qt.TransformationMode.FastTransformation)
        self.setAcceptedMouseButtons(Qt.MouseButton.LeftButton)
        self.setAcceptHoverEvents(True)

    def get_signals(self) -> ImageItemSignals:
        return self._signals

    def get_product(self) -> VisualizationProduct | None:
        return self._product

    def set_product(self, product: VisualizationProduct) -> None:
        image_rgba_f = product.get_image_rgba()
        # NOTE .copy() ensures imageRGBAi is not a view
        image_rgba_i = numpy.multiply(image_rgba_f, 255).astype(numpy.uint8).copy()

        try:
            image = QImage(
                image_rgba_i.data,
                image_rgba_i.shape[1],
                image_rgba_i.shape[0],
                image_rgba_i.strides[0],
                QImage.Format.Format_RGBA8888,
            )
            pixmap = QPixmap.fromImage(image)
        except Exception as exc:
            logger.exception(exc)
            pixmap = QPixmap()

        self._product = product
        self.setPixmap(pixmap)

    def clear_product(self) -> None:
        self._product = None
        self.setPixmap(QPixmap())

    def set_mouse_tool(self, mouse_tool: ImageMouseTool) -> None:
        self._fourier_item.setVisible(mouse_tool == ImageMouseTool.FOURIER_TOOL)
        self._mouse_tool = mouse_tool

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent | None) -> None:  # noqa: N802
        if event is None:
            return

        app = QApplication.instance()

        if app:
            cursor = Qt.CursorShape.CrossCursor

            if self._mouse_tool == ImageMouseTool.MOVE_TOOL:
                cursor = Qt.CursorShape.OpenHandCursor

            app.setOverrideCursor(cursor)  # type: ignore

        super().hoverEnterEvent(event)

    def hoverMoveEvent(self, event: QGraphicsSceneHoverEvent | None) -> None:  # noqa: N802
        if event is None:
            return

        pos = event.pos()

        if self._product is not None:
            info_text = self._product.get_info_text(pos.x(), pos.y())
            self._status_bar.showMessage(info_text)

        super().hoverMoveEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent | None) -> None:  # noqa: N802
        if event is None:
            return

        app = QApplication.instance()

        if app:
            app.restoreOverrideCursor()  # type: ignore

        self._status_bar.clearMessage()
        super().hoverLeaveEvent(event)

    def _change_override_cursor(self, cursor: Qt.CursorShape) -> None:
        app = QApplication.instance()

        if app:
            app.changeOverrideCursor(cursor)  # type: ignore

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent | None) -> None:  # noqa: N802
        if event is None:
            return

        match self._mouse_tool:
            case ImageMouseTool.MOVE_TOOL:
                self._change_override_cursor(Qt.CursorShape.ClosedHandCursor)
            case ImageMouseTool.RULER_TOOL:
                line = QLineF(event.pos(), event.pos())
                self.prepareGeometryChange()
                self._line_item.setLine(line)
                self._line_item.show()
            case ImageMouseTool.RECTANGLE_TOOL:
                self._rectangle_origin = event.pos()
                rect = QRectF(self._rectangle_origin, QSizeF())
                self.prepareGeometryChange()
                self._rectangle_item.setRect(rect)
                self._rectangle_item.show()
            case ImageMouseTool.LINE_CUT_TOOL:
                line = QLineF(event.pos(), event.pos())
                self.prepareGeometryChange()
                self._line_item.setLine(line)
                self._line_item.show()
            case ImageMouseTool.FOURIER_TOOL:
                self._fourier_origin = event.pos()
                rect = QRectF(self._fourier_origin, QSizeF())
                self.prepareGeometryChange()
                self._fourier_item.setRect(rect)
                self._fourier_item.show()

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent | None) -> None:  # noqa: N802
        if event is None:
            return

        match self._mouse_tool:
            case ImageMouseTool.MOVE_TOOL:
                self.setPos(self.scenePos() + event.scenePos() - event.lastScenePos())
            case ImageMouseTool.RULER_TOOL:
                origin = self._line_item.line().p1()
                line = QLineF(origin, event.pos())
                self.prepareGeometryChange()
                self._line_item.setLine(line)
                message1 = f'{line.length():.1f} pixels, {line.angle():.2f}\u00b0'
                message2 = f'{line.dx():.1f} \u00d7 {line.dy():.1f}'
                self._status_bar.showMessage(f'{message1} ({message2})')
            case ImageMouseTool.RECTANGLE_TOOL:
                rect = QRectF(self._rectangle_origin, event.pos()).normalized()
                center = rect.center()
                self.prepareGeometryChange()
                self._rectangle_item.setRect(rect)
                message1 = f'{rect.width():.1f} \u00d7 {rect.height():.1f}'
                message2 = f'{center.x():.1f}, {center.y():.1f}'
                self._status_bar.showMessage(f'Rectangle: {message1} (Center: {message2})')
            case ImageMouseTool.LINE_CUT_TOOL:
                origin = self._line_item.line().p1()
                line = QLineF(origin, event.pos())
                self.prepareGeometryChange()
                self._line_item.setLine(line)
                message1 = f'{line.length():.1f} pixels, {line.angle():.2f}\u00b0'
                message2 = f'{line.dx():.1f} \u00d7 {line.dy():.1f}'
                self._status_bar.showMessage(f'{message1} ({message2})')
            case ImageMouseTool.FOURIER_TOOL:
                rect = QRectF(self._fourier_origin, event.pos()).normalized()
                center = rect.center()
                self.prepareGeometryChange()
                self._fourier_item.setRect(rect)
                message1 = f'{rect.width():.1f} \u00d7 {rect.height():.1f}'
                message2 = f'{center.x():.1f}, {center.y():.1f}'
                self._status_bar.showMessage(f'Rectangle: {message1} (Center: {message2})')

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent | None) -> None:  # noqa: N802
        match self._mouse_tool:
            case ImageMouseTool.MOVE_TOOL:
                self._change_override_cursor(Qt.CursorShape.OpenHandCursor)
            case ImageMouseTool.RULER_TOOL:
                self._line_item.setLine(QLineF())
                self._line_item.hide()
            case ImageMouseTool.RECTANGLE_TOOL:
                self._signals.rectangle_finished.emit(self._rectangle_item.rect())
                self._rectangle_item.setRect(QRectF())
                self._rectangle_item.hide()
            case ImageMouseTool.LINE_CUT_TOOL:
                self._signals.line_cut_finished.emit(self._line_item.line())
                self._line_item.setLine(QLineF())
                self._line_item.hide()
            case ImageMouseTool.FOURIER_TOOL:
                self._signals.fourier_finished.emit(self._fourier_item.rect())


class LineCutDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.figure = Figure()
        self.figure_canvas = FigureCanvasQTAgg(self.figure)
        self.navigation_toolbar = NavigationToolbar(self.figure_canvas, self)
        self.axes = self.figure.add_subplot(111)

        layout = QVBoxLayout()
        layout.addWidget(self.navigation_toolbar)
        layout.addWidget(self.figure_canvas)
        self.setLayout(layout)

        self.setWindowTitle('Line-Cut Dialog')


class RectangleView(QGroupBox):
    @staticmethod
    def _create_read_only_line_edit() -> QLineEdit:
        line_edit = QLineEdit()

        palette = line_edit.palette()
        palette.setColor(QPalette.ColorRole.Base, palette.color(QPalette.ColorRole.Window))
        palette.setColor(QPalette.ColorRole.Text, palette.color(QPalette.ColorRole.WindowText))
        line_edit.setPalette(palette)

        line_edit.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        line_edit.setReadOnly(True)

        return line_edit

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Rectangle', parent)
        self.center_x_line_edit = RectangleView._create_read_only_line_edit()
        self.center_y_line_edit = RectangleView._create_read_only_line_edit()
        self.width_line_edit = RectangleView._create_read_only_line_edit()
        self.height_line_edit = RectangleView._create_read_only_line_edit()

        layout = QFormLayout()
        layout.addRow('Center X:', self.center_x_line_edit)
        layout.addRow('Center Y:', self.center_y_line_edit)
        layout.addRow('Width:', self.width_line_edit)
        layout.addRow('Height:', self.height_line_edit)
        self.setLayout(layout)


class HistogramDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.figure = Figure()
        self.figure_canvas = FigureCanvasQTAgg(self.figure)
        self.navigation_toolbar = NavigationToolbar(self.figure_canvas, self)
        self.axes = self.figure.add_subplot(111)
        self.rectangle_view = RectangleView()

        layout = QVBoxLayout()
        layout.addWidget(self.navigation_toolbar)
        layout.addWidget(self.figure_canvas, 1)
        layout.addWidget(self.rectangle_view)
        self.setLayout(layout)

        self.setWindowTitle('Histogram')


class VisualizationView(QGraphicsView):
    def wheelEvent(self, event: QWheelEvent | None) -> None:  # noqa: N802
        if event is None:
            return

        old_position = self.mapToScene(event.pos())

        zoom_base = 1.25
        zoom = zoom_base if event.angleDelta().y() > 0 else 1.0 / zoom_base
        self.scale(zoom, zoom)

        new_position = self.mapToScene(event.pos())

        delta_position = new_position - old_position
        self.translate(delta_position.x(), delta_position.y())


class VisualizationWidget(QGroupBox):
    def __init__(self, title: str, parent: QWidget | None = None) -> None:
        super().__init__(title, parent)
        self.tool_bar = QToolBar('Tools')
        self.home_action = QAction(QIcon(':/icons/home'), 'Home')
        self.save_action = QAction(QIcon(':/icons/save'), 'Save Image')
        self.autoscale_action = QAction(QIcon(':/icons/autoscale'), 'Autoscale Color Axis')

        self.move_action = QAction(QIcon(':/icons/move'), 'Move')
        self.ruler_action = QAction(QIcon(':/icons/ruler'), 'Ruler')
        self.rectangle_action = QAction(QIcon(':/icons/rectangle'), 'Rectangle')
        self.line_cut_action = QAction(QIcon(':/icons/line-cut'), 'Line-Cut Profile')

        self.visualization_view = VisualizationView()

        self.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        self.tool_bar.setFloatable(False)
        self.tool_bar.setIconSize(QSize(32, 32))
        self.tool_bar.addAction(self.home_action)
        self.tool_bar.addAction(self.save_action)
        self.tool_bar.addAction(self.autoscale_action)
        self.tool_bar.addSeparator()
        self.tool_bar.addAction(self.move_action)
        self.tool_bar.addAction(self.ruler_action)
        self.tool_bar.addAction(self.rectangle_action)
        self.tool_bar.addAction(self.line_cut_action)
        self.tool_bar.addSeparator()

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.tool_bar)
        layout.addWidget(self.visualization_view)
        self.setLayout(layout)


class VisualizationParametersView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Visualization', parent)
        self.renderer_combo_box = QComboBox()
        self.transformation_combo_box = QComboBox()
        self.variant_combo_box = QComboBox()
        self.min_display_value_line_edit = DecimalLineEdit.create_instance()
        self.max_display_value_line_edit = DecimalLineEdit.create_instance()

        layout = QFormLayout()
        layout.addRow('Renderer:', self.renderer_combo_box)
        layout.addRow('Transform:', self.transformation_combo_box)
        layout.addRow('Variant:', self.variant_combo_box)
        layout.addRow('Min Display Value:', self.min_display_value_line_edit)
        layout.addRow('Max Display Value:', self.max_display_value_line_edit)
        self.setLayout(layout)



"""Reconstructor interfaces, I/O data containers, and assembled diffraction data management."""

from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator, Sequence
from dataclasses import dataclass, field
from enum import auto, Enum
from pathlib import Path
import logging

import numpy

from .common import BYTES_PER_MEGABYTE
from .diffraction import (
    BadPixels,
    DiffractionIndexes,
    DiffractionPattern,
    DiffractionPatternCounts,
    DiffractionPatternDType,
    DiffractionPatterns,
)
from .probe_positions import ProbePositionSequence, ProbePosition
from .product import LossValue, Product

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ReconstructInput:
    """All data required to start a reconstruction: patterns, bad-pixel mask, and initial product."""

    diffraction_patterns: DiffractionPatterns
    bad_pixels: BadPixels
    product: Product


@dataclass(frozen=True)
class ReconstructOutput:
    """Reconstruction result yielded at each iteration: updated product, progress, and status code."""

    product: Product
    progress: int = 0
    status: int = 0


class Reconstructor(ABC):
    """Abstract interface for iterative ptychography reconstruction algorithms."""

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def get_progress_goal(self) -> int:
        pass

    @abstractmethod
    def reconstruct(self, parameters: ReconstructInput) -> Iterator[ReconstructOutput]:
        pass


@dataclass(frozen=True)
class TrainOutput:
    """Training result yielded at each step: loss curves, progress, and status code."""

    training_loss: Sequence[LossValue] = field(default_factory=list)
    validation_loss: Sequence[LossValue] = field(default_factory=list)
    progress: int = 0
    status: int = 0


class TrainableReconstructor(Reconstructor):
    """Reconstructor that also supports ML model loading and training."""

    @abstractmethod
    def is_model_loaded(self) -> bool:
        pass

    @abstractmethod
    def get_model_file_filter(self) -> str:
        pass

    @abstractmethod
    def load_model_from_file(self, file_path: Path) -> None:
        pass

    @abstractmethod
    def get_training_data_file_filter(self) -> str:
        pass

    @abstractmethod
    def export_training_data(self, file_path: Path, parameters: ReconstructInput) -> None:
        pass

    @abstractmethod
    def train(self, input_path: Path, output_path: Path) -> Iterator[TrainOutput]:
        pass


class NullReconstructor(TrainableReconstructor):
    """No-op TrainableReconstructor used as a placeholder when no algorithm is available."""

    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def get_progress_goal(self) -> int:
        return 0

    def reconstruct(self, parameters: ReconstructInput) -> Iterator[ReconstructOutput]:
        yield from ()

    def is_model_loaded(self) -> bool:
        return False

    def get_model_file_filter(self) -> str:
        return str()

    def load_model_from_file(self, file_path: Path) -> None:
        pass

    def get_training_data_file_filter(self) -> str:
        return str()

    def export_training_data(self, file_path: Path, parameters: ReconstructInput) -> None:
        pass

    def train(self, input_path: Path, output_path: Path) -> Iterator[TrainOutput]:
        yield from ()


class ReconstructorLibrary(Iterable[Reconstructor], ABC):
    """Iterable collection of Reconstructor instances provided by a plugin library."""

    def __init__(self, logger_name: str) -> None:
        super().__init__()
        self._logger = logging.getLogger(logger_name)

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    def log_levels(self) -> Iterable[str]:
        return ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG')

    def get_logger(self) -> logging.Logger:
        return self._logger

    def get_log_level(self) -> str:
        level = self._logger.getEffectiveLevel()
        return logging.getLevelName(level)

    def set_log_level(self, name: str) -> None:
        name_before = self.get_log_level()

        try:
            self._logger.setLevel(name)
        except ValueError:
            self._logger.error(f'Bad log level "{name}".')

        name_after = self.get_log_level()
        self._logger.info(f'Changed {self.name} logging level {name_before} -> {name_after}')


class PositionIndexFilter(Enum):
    """Filter scan points by scan index."""

    ALL = auto()
    ODD = auto()
    EVEN = auto()

    def __call__(self, index: int) -> bool:
        """Return True to include the scan point, False to exclude it."""
        if self is PositionIndexFilter.ODD:
            return index & 1 != 0
        elif self is PositionIndexFilter.EVEN:
            return index & 1 == 0

        return True


class AssembledDiffractionData:
    """In-memory store for a complete set of indexed diffraction patterns and their bad-pixel mask."""

    def __init__(
        self, indexes: DiffractionIndexes, patterns: DiffractionPatterns, bad_pixels: BadPixels
    ) -> None:
        self._indexes = indexes
        self._patterns = patterns
        self._bad_pixels = bad_pixels

        if indexes.ndim != 1:
            raise ValueError(
                f'Unexpected number of dimensions for indexes! (actual={indexes.ndim} expected=1)'
            )

        if patterns.ndim != 3:
            raise ValueError(
                f'Unexpected number of dimensions for patterns! (actual={patterns.ndim} expected=3)'
            )

        if bad_pixels.ndim != 2:
            raise ValueError(
                f'Unexpected number of dimensions for bad pixels! (actual={bad_pixels.ndim} expected=2)'
            )

        if indexes.shape[0] != patterns.shape[0]:
            raise ValueError('Number of indexes does not match number of patterns!')

        if patterns.shape[1:] != bad_pixels.shape:
            raise ValueError('Patterns shape does not match bad pixels shape!')

    @classmethod
    def create_null(cls) -> AssembledDiffractionData:
        return cls(
            indexes=numpy.zeros(1, dtype=int),
            patterns=numpy.zeros((1, 1, 1), dtype=int),
            bad_pixels=numpy.zeros((1, 1), dtype=bool),
        )

    def get_patterns_shape(self) -> tuple[int, int, int]:
        return self._patterns.shape

    def get_patterns_dtype(self) -> DiffractionPatternDType:
        return self._patterns.dtype

    def get_pattern(self, index: int) -> DiffractionPattern:
        return self._patterns[index]

    def get_bad_pixels(self) -> BadPixels:
        return self._bad_pixels

    def assemble(self, data: AssembledDiffractionData, offset: int) -> AssembledDiffractionData:
        assembled_indexes = slice(offset, offset + len(data._indexes))

        self._indexes[assembled_indexes] = data._indexes
        indexes_view = self._indexes[assembled_indexes]
        indexes_view.flags.writeable = False

        self._patterns[assembled_indexes, :, :] = data._patterns
        patterns_view = self._patterns[assembled_indexes, :, :]
        patterns_view.flags.writeable = False

        return AssembledDiffractionData(
            indexes=indexes_view,
            patterns=patterns_view,
            bad_pixels=data._bad_pixels,
        )

    def get_indexes(self) -> DiffractionIndexes:
        return self._indexes[self._indexes >= 0]

    def get_patterns(self) -> DiffractionPatterns:
        return self._patterns[self._indexes >= 0]

    def get_pattern_counts(self) -> DiffractionPatternCounts:
        good_pixels = numpy.logical_not(self._bad_pixels)
        assembled_patterns = self.get_patterns()
        pattern_counts = numpy.sum(assembled_patterns[:, good_pixels], axis=-1)
        return pattern_counts

    def get_average_pattern(self) -> DiffractionPattern:
        assembled_patterns = self.get_patterns()
        return numpy.mean(assembled_patterns, axis=0)

    def prepare_reconstruct_input(
        self,
        product: Product,
        index_filter: PositionIndexFilter = PositionIndexFilter.ALL,
    ) -> ReconstructInput:
        pattern_indexes = [int(index) for index in self.get_indexes()]
        logger.debug(f'{pattern_indexes=}')
        position_indexes = [
            int(point.index) for point in product.probe_positions if index_filter(point.index)
        ]
        logger.debug(f'{position_indexes=}')
        common_indexes = sorted(set(pattern_indexes).intersection(position_indexes))
        logger.debug(f'{common_indexes=}')

        patterns = numpy.take(
            self.get_patterns(),
            common_indexes,
            axis=0,
        )

        point_list: list[ProbePosition] = list()
        point_iterator = iter(product.probe_positions)

        for index in common_indexes:
            while True:
                point = next(point_iterator)

                if point.index == index:
                    point_list.append(point)
                    break

        product = Product(
            metadata=product.metadata,
            probe_positions=ProbePositionSequence(point_list),
            probes=product.probes,  # TODO remap if needed
            object_=product.object_,
            losses=product.losses,
        )

        return ReconstructInput(patterns, self._bad_pixels, product)

    def __str__(self) -> str:
        number, height, width = self._patterns.shape
        dtype = str(self._patterns.dtype)
        size_MB = self._patterns.nbytes / BYTES_PER_MEGABYTE  # noqa: N806
        return f'{number} x {height}H x {width}W {dtype} [{size_MB:.2f}MB]'



from pathlib import Path

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class ProductSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Products')
        self._group.add_observer(self)

        self.name = self._group.create_string_parameter('Name', 'Unnamed')
        self.file_path = self._group.create_path_parameter('FilePath', Path('/path/to/product.h5'))
        self.file_type = self._group.create_string_parameter('FileType', 'HDF5')
        self.detector_distance_m = self._group.create_real_parameter(
            'DetectorDistanceInMeters', 1.0, minimum=0.0
        )
        self.probe_energy_eV = self._group.create_real_parameter(
            'ProbeEnergyInElectronVolts', 10000.0, minimum=0.0
        )
        self.probe_photon_count = self._group.create_real_parameter(
            'ProbePhotonCount', 0.0, minimum=0.0
        )
        self.exposure_time_s = self._group.create_real_parameter(
            'ExposureTimeInSeconds', 0.0, minimum=0.0
        )
        self.mass_attenuation_m2_kg = self._group.create_real_parameter(
            'MassAttenuationSquareMetersPerKilogram', 0.0, minimum=0.0
        )
        self.tomography_angle_deg = self._group.create_real_parameter(
            'TomographyAngleInDegrees', 0.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from __future__ import annotations
from collections.abc import Sequence
import logging

from PyQt5.QtCore import Qt, QModelIndex
from PyQt5.QtWidgets import (
    QButtonGroup,
    QComboBox,
    QHBoxLayout,
    QRadioButton,
    QWidget,
)

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.product import LossValue

from ...model.processing import ProcessingProgressMonitor
from ...model.product import ProductRepository, ProductRepositoryItem, ProductRepositoryObserver
from ...model.product.metadata import MetadataRepositoryItem
from ...model.product.object import ObjectRepositoryItem
from ...model.product.probe import ProbeRepositoryItem
from ...model.product.probe_positions import ProbePositionsRepositoryItem
from ...view.processing import ProcessingStatusView
from ..parametric import ParameterViewController
from ..product.list_model import ProductRepositoryListModel

logger = logging.getLogger(__name__)


class ProcessingStatusController(Observer):
    def __init__(
        self,
        product_repository: ProductRepository,
        monitor: ProcessingProgressMonitor,
        view: ProcessingStatusView,
    ) -> None:
        super().__init__()
        self._product_repository = product_repository
        self._monitor = monitor
        self._view = view
        self._view.text_edit.setReadOnly(True)

        view.stop_button.clicked.connect(monitor.stop_processing)

        self._sync_model_to_view()
        monitor.add_observer(self)

    def plot_losses(self, product_index: int) -> None:
        # TODO fix so that plot and progress cannot mismatch
        if product_index < 0:
            self._view.axes.clear()
            return

        try:
            item = self._product_repository[product_index]
        except IndexError as exc:
            logger.exception(exc)
            return

        epoch = [loss.epoch for loss in item.get_losses()]
        losses = [loss.value for loss in item.get_losses()]

        ax = self._view.axes
        ax.clear()
        ax.set_xlabel('Epoch')
        ax.set_ylabel('Loss')
        ax.grid(True)
        ax.plot(epoch, losses, '.-', label='Loss', linewidth=1.5)
        self._view.figure_canvas.draw()

    def _sync_model_to_view(self) -> None:
        log_handler = self._monitor.get_log_handler()

        for text in log_handler.messages():
            self._view.text_edit.appendPlainText(text)

        progress_goal = self._monitor.get_progress_goal()
        progress_bar = self._view.progress_bar

        if self._monitor.is_processing and progress_goal > 0:
            progress_bar.show()
            progress_bar.setRange(0, progress_goal)
            progress_bar.setValue(self._monitor.get_progress())
            self._view.stop_button.show()
        else:
            progress_bar.hide()
            self._view.stop_button.hide()

    def _update(self, observable: Observable) -> None:
        if observable is self._monitor:
            self._sync_model_to_view()


class ProductParameterViewController(ParameterViewController, ProductRepositoryObserver):
    def __init__(
        self,
        repository: ProductRepository,
        status_controller: ProcessingStatusController,
        *,
        tool_tip: str = '',
    ) -> None:
        super().__init__()
        self._repository = repository
        self._status_controller = status_controller
        self._model = ProductRepositoryListModel(repository)
        self._widget = QComboBox()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self._widget.setModel(self._model)
        self._widget.currentIndexChanged.connect(status_controller.plot_losses)

        repository.add_observer(self)

    def get_widget(self) -> QComboBox:
        return self._widget

    def handle_item_inserted(self, index: int, item: ProductRepositoryItem) -> None:
        parent = QModelIndex()
        self._model.beginInsertRows(parent, index, index)
        self._model.endInsertRows()

    def handle_metadata_changed(self, index: int, item: MetadataRepositoryItem) -> None:
        top_left = self._model.index(index, 0)
        bottom_right = self._model.index(index, 0)
        self._model.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.DisplayRole])

    def handle_probe_positions_changed(
        self, index: int, item: ProbePositionsRepositoryItem
    ) -> None:
        pass

    def handle_probe_changed(self, index: int, item: ProbeRepositoryItem) -> None:
        pass

    def handle_object_changed(self, index: int, item: ObjectRepositoryItem) -> None:
        pass

    def handle_losses_changed(self, index: int, losses: Sequence[LossValue]) -> None:
        current_index = self._widget.currentIndex()

        if index == current_index:
            self._status_controller.plot_losses(index)

    def handle_item_removed(self, index: int, item: ProductRepositoryItem) -> None:
        parent = QModelIndex()
        self._model.beginRemoveRows(parent, index, index)
        self._model.endRemoveRows()


class ComputeParameterViewController(ParameterViewController):
    def __init__(self, *, tool_tip: str = '') -> None:
        self._local_button = QRadioButton('Local')
        self._remote_button = QRadioButton('Remote')
        self._button_group = QButtonGroup()
        self._widget = QWidget()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self._button_group.addButton(self._local_button)
        self._button_group.addButton(self._remote_button)
        self._button_group.setExclusive(True)
        self._local_button.setChecked(True)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._local_button)
        layout.addWidget(self._remote_button)
        layout.addStretch()
        self._widget.setLayout(layout)

    def is_computing_local(self) -> bool:
        return self._button_group.checkedButton() is self._local_button

    def get_widget(self) -> QWidget:
        return self._widget



from __future__ import annotations

from PyQt5.QtWidgets import QDialogButtonBox, QListView, QVBoxLayout, QWidget


class SettingsView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.list_view = QListView()
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Open | QDialogButtonBox.StandardButton.Save
        )

        layout = QVBoxLayout()
        layout.addWidget(self.list_view)
        layout.addWidget(self.button_box)
        self.setLayout(layout)



from scipy.interpolate import griddata, RBFInterpolator
import numpy

from ptychodus.api.fluorescence import ElementMap, UpscalingStrategy
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.product import Product


class IdentityUpscaling(UpscalingStrategy):
    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        return emap


class GridDataUpscaling(UpscalingStrategy):
    def __init__(self, method: str) -> None:
        self._method = method

    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        object_geometry = product.object_.get_geometry()
        scan_coords_px: list[float] = list()

        for scan_point in product.probe_positions:
            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            scan_coords_px.append(object_point.coordinate_y_px)
            scan_coords_px.append(object_point.coordinate_x_px)

        points = numpy.reshape(scan_coords_px, (-1, 2))
        values = emap.counts_per_second.flat
        YY, XX = numpy.mgrid[: object_geometry.height_px, : object_geometry.width_px]  # noqa: N806
        query_points = numpy.transpose((YY.flat, XX.flat))

        cps = griddata(points, values, query_points, method=self._method, fill_value=0.0).reshape(
            XX.shape
        )

        return ElementMap(emap.name, cps.astype(emap.counts_per_second.dtype))


class RadialBasisFunctionUpscaling(UpscalingStrategy):
    def __init__(
        self,
        kernel: str,
        *,
        neighbors: int | None = 25,
        epsilon: float | None = None,
        degree: int | None = None,
    ) -> None:
        self._kernel = kernel
        self._neighbors = neighbors
        self._epsilon = epsilon
        self._degree = degree

    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        object_geometry = product.object_.get_geometry()
        scan_coords_px: list[float] = list()

        for scan_point in product.probe_positions:
            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            scan_coords_px.append(object_point.coordinate_y_px)
            scan_coords_px.append(object_point.coordinate_x_px)

        interpolator = RBFInterpolator(
            numpy.reshape(scan_coords_px, (-1, 2)),
            emap.counts_per_second.flat,
            kernel=self._kernel,
            neighbors=self._neighbors,
            epsilon=self._epsilon,
            degree=self._degree,
        )
        YY, XX = numpy.mgrid[: object_geometry.height_px, : object_geometry.width_px]  # noqa: N806
        cps = interpolator(numpy.transpose((YY.flat, XX.flat)))
        return ElementMap(emap.name, cps.astype(emap.counts_per_second.dtype).reshape(XX.shape))


def register_plugins(registry: PluginRegistry) -> None:
    # TODO natural neighbor
    # TODO kriging
    # TODO inverse distance weighting

    registry.upscaling_strategies.register_plugin(
        IdentityUpscaling(),
        display_name='Identity',
    )
    registry.upscaling_strategies.register_plugin(
        GridDataUpscaling('nearest'),
        display_name='Nearest Neighbor',
    )
    registry.upscaling_strategies.register_plugin(
        GridDataUpscaling('linear'),
        display_name='Linear',
    )
    registry.upscaling_strategies.register_plugin(
        GridDataUpscaling('cubic'),
        display_name='Cubic',
    )
    registry.upscaling_strategies.register_plugin(
        RadialBasisFunctionUpscaling('linear'),
        display_name='Linear RBF',
    )
    registry.upscaling_strategies.register_plugin(
        RadialBasisFunctionUpscaling('thin_plate_spline'),
        display_name='Thin Plate Spline RBF',
    )
    registry.upscaling_strategies.register_plugin(
        RadialBasisFunctionUpscaling('cubic'),
        display_name='Cubic RBF',
    )
    registry.upscaling_strategies.register_plugin(
        RadialBasisFunctionUpscaling('quintic'),
        display_name='Quintic RBF',
    )
    registry.upscaling_strategies.register_plugin(
        RadialBasisFunctionUpscaling('multiquadric'),
        display_name='Multiquadric RBF',
    )
    registry.upscaling_strategies.register_plugin(
        RadialBasisFunctionUpscaling('inverse_multiquadric'),
        display_name='Inverse Multiquadric RBF',
    )
    registry.upscaling_strategies.register_plugin(
        RadialBasisFunctionUpscaling('inverse_quadratic'),
        display_name='Inverse Quadratic RBF',
    )
    registry.upscaling_strategies.register_plugin(
        RadialBasisFunctionUpscaling('gaussian'),
        display_name='Gaussian RBF',
    )



from __future__ import annotations
from abc import abstractmethod, ABC
import logging

from ptychodus.api.parametric import Parameter, ParameterGroup
from ptychodus.api.product import ProductMetadata

from .settings import ProductSettings

logger = logging.getLogger(__name__)


class UniqueNameFactory(ABC):
    @abstractmethod
    def create_unique_name(self, candidate_name: str) -> str:
        pass


class UniqueStringParameter(Parameter[str]):
    def __init__(
        self, value: str | None, name_factory: UniqueNameFactory, parent: Parameter[str]
    ) -> None:
        super().__init__(parent)
        self._value = name_factory.create_unique_name(value or parent.get_value())
        self._name_factory = name_factory

    def get_value(self) -> str:
        return self._value

    def set_value(self, value: str, *, notify: bool = True) -> None:
        if value:
            if self._value != value:
                self._value = self._name_factory.create_unique_name(value)

                if notify:
                    self.notify_observers()
        else:
            self.notify_observers()

    def get_value_as_string(self) -> str:
        return str(self._value)

    def set_value_from_string(self, value: str) -> None:
        self.set_value(str(value))

    def copy(self) -> UniqueStringParameter:
        return UniqueStringParameter(self.get_value(), self._name_factory, self)


class MetadataRepositoryItem(ParameterGroup):
    def __init__(
        self,
        settings: ProductSettings,
        name_factory: UniqueNameFactory,
        *,
        name: str = '',
        comments: str = '',
        detector_distance_m: float | None = None,
        probe_energy_eV: float | None = None,  # noqa: N803
        probe_photon_count: float | None = None,
        exposure_time_s: float | None = None,
        mass_attenuation_m2_kg: float | None = None,
        tomography_angle_deg: float | None = None,
    ) -> None:
        super().__init__()
        self._settings = settings

        self.name = UniqueStringParameter(name, name_factory, settings.name.copy())
        self._add_parameter('name', self.name)

        self.comments = self.create_string_parameter('comments', comments)

        self.detector_distance_m = settings.detector_distance_m.copy()

        if detector_distance_m is not None:
            self.detector_distance_m.set_value(detector_distance_m)

        self._add_parameter('detector_distance_m', self.detector_distance_m)

        self.probe_energy_eV = settings.probe_energy_eV.copy()

        if probe_energy_eV is not None:
            self.probe_energy_eV.set_value(probe_energy_eV)

        self._add_parameter('probe_energy_eV', self.probe_energy_eV)

        self.probe_photon_count = settings.probe_photon_count.copy()

        if probe_photon_count is not None:
            self.probe_photon_count.set_value(probe_photon_count)

        self._add_parameter('probe_photon_count', self.probe_photon_count)

        self.exposure_time_s = settings.exposure_time_s.copy()

        if exposure_time_s is not None:
            self.exposure_time_s.set_value(exposure_time_s)

        self._add_parameter('exposure_time_s', self.exposure_time_s)

        self.mass_attenuation_m2_kg = settings.mass_attenuation_m2_kg.copy()

        if mass_attenuation_m2_kg is not None:
            self.mass_attenuation_m2_kg.set_value(mass_attenuation_m2_kg)

        self._add_parameter('mass_attenuation_m2_kg', self.mass_attenuation_m2_kg)

        self.tomography_angle_deg = settings.tomography_angle_deg.copy()

        if tomography_angle_deg is not None:
            self.tomography_angle_deg.set_value(tomography_angle_deg)

        self._add_parameter('tomography_angle_deg', self.tomography_angle_deg)

    def assign(self, metadata: ProductMetadata) -> None:
        self.name.set_value(metadata.name)
        self.comments.set_value(metadata.comments)
        self.detector_distance_m.set_value(metadata.detector_distance_m)
        self.probe_energy_eV.set_value(metadata.probe_energy_eV)
        self.probe_photon_count.set_value(metadata.probe_photon_count)
        self.exposure_time_s.set_value(metadata.exposure_time_s)
        self.mass_attenuation_m2_kg.set_value(metadata.mass_attenuation_m2_kg)
        self.tomography_angle_deg.set_value(metadata.tomography_angle_deg)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    def get_metadata(self) -> ProductMetadata:
        return ProductMetadata(
            name=self.name.get_value(),
            comments=self.comments.get_value(),
            detector_distance_m=self.detector_distance_m.get_value(),
            probe_energy_eV=self.probe_energy_eV.get_value(),
            probe_photon_count=self.probe_photon_count.get_value(),
            exposure_time_s=self.exposure_time_s.get_value(),
            mass_attenuation_m2_kg=self.mass_attenuation_m2_kg.get_value(),
            tomography_angle_deg=self.tomography_angle_deg.get_value(),
        )



from __future__ import annotations
from collections.abc import Sequence
import logging

import numpy

from ptychodus.api.common import RealArrayType, TWO_PI_J, lerp
from ptychodus.api.object import Object, ObjectGeometryProvider

from .builder import ObjectBuilder
from .settings import ObjectSettings

logger = logging.getLogger(__name__)


class DeadLeavesObjectBuilder(ObjectBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ObjectSettings) -> None:
        super().__init__(settings, 'dead_leaves')
        self._rng = rng
        self._settings = settings

        self.leaf_radius_lower_px = settings.leaf_radius_lower_px.copy()
        self._add_parameter('leaf_radius_lower_px', self.leaf_radius_lower_px)
        self.leaf_radius_upper_px = settings.leaf_radius_upper_px.copy()
        self._add_parameter('leaf_radius_upper_px', self.leaf_radius_upper_px)
        self.leaf_radius_power_law_exponent = settings.leaf_radius_power_law_exponent.copy()
        self._add_parameter('leaf_radius_power_law_exponent', self.leaf_radius_power_law_exponent)

        self.leaf_amplitude_lower = settings.leaf_amplitude_lower.copy()
        self._add_parameter('leaf_amplitude_lower', self.leaf_amplitude_lower)
        self.leaf_amplitude_upper = settings.leaf_amplitude_upper.copy()
        self._add_parameter('leaf_amplitude_upper', self.leaf_amplitude_upper)

        self.leaf_phase_lower_turns = settings.leaf_phase_lower_turns.copy()
        self._add_parameter('leaf_phase_lower_turns', self.leaf_phase_lower_turns)
        self.leaf_phase_upper_turns = settings.leaf_phase_upper_turns.copy()
        self._add_parameter('leaf_phase_upper_turns', self.leaf_phase_upper_turns)

        # TODO consider using Poisson disk sampling or CVT for sample positions
        # TODO include pixel aspect ratio in sample position calculation
        self._sample_positions = [
            (1.0 / 3, 1.0 / 3),
            (1.0 / 3, 2.0 / 3),
            (2.0 / 3, 1.0 / 3),
            (2.0 / 3, 2.0 / 3),
        ]

    def copy(self) -> DeadLeavesObjectBuilder:
        builder = DeadLeavesObjectBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def _sample_leaf_radius_px(self) -> float:
        rmin_px = self.leaf_radius_lower_px.get_value()
        rmax_px = self.leaf_radius_upper_px.get_value()
        beta = 1.0 - self.leaf_radius_power_law_exponent.get_value()
        coef = numpy.power(rmax_px / rmin_px, beta) - 1.0
        return rmin_px * numpy.power(1.0 + self._rng.uniform() * coef, 1.0 / beta)

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        if self.leaf_radius_lower_px.get_value() >= self.leaf_radius_upper_px.get_value():
            raise ValueError('leaf_radius_lower_px must be less than leaf_radius_upper_px')

        if self.leaf_amplitude_lower.get_value() >= self.leaf_amplitude_upper.get_value():
            raise ValueError('leaf_amplitude_lower must be less than leaf_amplitude_upper')

        if self.leaf_phase_lower_turns.get_value() >= self.leaf_phase_upper_turns.get_value():
            raise ValueError('leaf_phase_lower_turns must be less than leaf_phase_upper_turns')

        geometry = geometry_provider.get_object_geometry()
        object_shape = (geometry.height_px, geometry.width_px)
        is_covered = numpy.zeros(object_shape, dtype=bool)
        amplitude: RealArrayType = numpy.zeros(object_shape, dtype=float)
        phase_turns: RealArrayType = numpy.zeros(object_shape, dtype=float)

        position_y_px, position_x_px = numpy.indices(object_shape)
        num_covered_pixels = 0

        for leaf in range(100000):  # large value eliminates possiblity of infinite loop
            leaf_radius_px = self._sample_leaf_radius_px()
            leaf_position_x_px = self._rng.uniform(
                -leaf_radius_px, geometry.width_px + leaf_radius_px
            )
            leaf_position_y_px = self._rng.uniform(
                -leaf_radius_px, geometry.height_px + leaf_radius_px
            )
            leaf_phase_turns = self._rng.uniform(
                self.leaf_phase_lower_turns.get_value(),
                self.leaf_phase_upper_turns.get_value(),
                size=(1, 1),
            )
            leaf_amplitude = numpy.sqrt(
                self._rng.uniform(
                    self.leaf_amplitude_lower.get_value() ** 2,
                    self.leaf_amplitude_upper.get_value() ** 2,
                    size=(1, 1),
                )
            )

            leaf_counts = numpy.zeros_like(is_covered, dtype=int)
            dx = position_x_px - leaf_position_x_px
            dy = position_y_px - leaf_position_y_px

            # multi-sample anti-aliasing
            for u_y, u_x in self._sample_positions:
                sample_distance_px = numpy.hypot(dy + u_y, dx + u_x)
                leaf_counts[sample_distance_px <= leaf_radius_px] += 1

            num_samples = len(self._sample_positions)
            leaf_coverage = leaf_counts / num_samples
            amplitude = lerp(amplitude, leaf_amplitude, leaf_coverage)
            phase_turns = lerp(phase_turns, leaf_phase_turns, leaf_coverage)

            is_covered |= leaf_counts == num_samples
            num_covered_pixels = numpy.count_nonzero(is_covered).item()

            covered_pct = 100 * num_covered_pixels / is_covered.size
            logger.info(f'leaves = {leaf}, covered = {covered_pct:.2f}%')

            if num_covered_pixels == is_covered.size:
                break

        array = amplitude * numpy.exp(TWO_PI_J * phase_turns)

        return self._create_object(
            array=array.astype('complex'),
            layer_spacing_m=layer_spacing_m,
            pixel_geometry=geometry.get_pixel_geometry(),
            center=geometry.get_center(),
        )



from __future__ import annotations
from dataclasses import dataclass

import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    BadPixels,
    CropCenter,
    DiffractionArray,
    DiffractionPatterns,
    SimpleDiffractionArray,
)


@dataclass(frozen=True)
class DiffractionPatternFilterValues:
    lower_bound: int | None
    upper_bound: int | None

    def apply(self, data: DiffractionPatterns) -> DiffractionPatterns:
        if self.lower_bound is not None:
            data[data < self.lower_bound] = 0

        if self.upper_bound is not None:
            data[data >= self.upper_bound] = 0

        return data


class DiffractionPatternCrop:
    def __init__(self, center: CropCenter, extent: ImageExtent) -> None:
        center_x = center.position_x_px
        radius_x = extent.width_px // 2
        self.slice_x = slice(center_x - radius_x, center_x + radius_x)

        center_y = center.position_y_px
        radius_y = extent.height_px // 2
        self.slice_y = slice(center_y - radius_y, center_y + radius_y)

    def apply_bool(self, data: BadPixels) -> BadPixels:
        return data[self.slice_y, self.slice_x]

    def apply(self, data: DiffractionPatterns) -> DiffractionPatterns:
        return data[:, self.slice_y, self.slice_x]


@dataclass(frozen=True)
class DiffractionPatternBinning:
    bin_size_x: int
    bin_size_y: int

    def apply_bool(self, data: BadPixels) -> BadPixels:
        binned_width = data.shape[-1] // self.bin_size_x
        binned_height = data.shape[-2] // self.bin_size_y
        shape = (binned_height, self.bin_size_y, binned_width, self.bin_size_x)
        return numpy.logical_and.reduce(data.reshape(shape), axis=(-3, -1), keepdims=False)

    def apply(self, data: DiffractionPatterns) -> DiffractionPatterns:
        binned_width = data.shape[-1] // self.bin_size_x
        binned_height = data.shape[-2] // self.bin_size_y
        shape = (-1, binned_height, self.bin_size_y, binned_width, self.bin_size_x)
        return numpy.sum(data.reshape(shape), axis=(-3, -1), keepdims=False)


@dataclass(frozen=True)
class DiffractionPatternPadding:
    pad_x: int
    pad_y: int

    def apply_bool(self, data: BadPixels) -> BadPixels:
        pad_width = (self.pad_y, self.pad_y, self.pad_x, self.pad_x)
        return numpy.pad(data, pad_width, mode='constant', constant_values=False)

    def apply(self, data: DiffractionPatterns) -> DiffractionPatterns:
        pad_width = (0, 0, self.pad_y, self.pad_y, self.pad_x, self.pad_x)
        return numpy.pad(data, pad_width, mode='constant', constant_values=0)


@dataclass(frozen=True)
class DiffractionPatternProcessor:
    crop: DiffractionPatternCrop | None
    filter_values: DiffractionPatternFilterValues | None
    binning: DiffractionPatternBinning | None
    padding: DiffractionPatternPadding | None
    hflip: bool
    vflip: bool
    transpose: bool

    def process_bad_pixels(self, bad_pixels: BadPixels) -> BadPixels:
        if bad_pixels.ndim != 2:
            raise ValueError(f'Invalid bad_pixel dimensions! (shape={bad_pixels.shape})')

        processed_bad_pixels = bad_pixels.copy()

        if self.crop is not None:
            processed_bad_pixels = self.crop.apply_bool(processed_bad_pixels)

        if self.binning is not None:
            processed_bad_pixels = self.binning.apply_bool(processed_bad_pixels)

        if self.padding is not None:
            processed_bad_pixels = self.padding.apply_bool(processed_bad_pixels)

        if self.hflip:
            processed_bad_pixels = numpy.flip(processed_bad_pixels, axis=-1)

        if self.vflip:
            processed_bad_pixels = numpy.flip(processed_bad_pixels, axis=-2)

        if self.transpose:
            processed_bad_pixels = numpy.transpose(processed_bad_pixels, axes=(0, 2, 1))

        return processed_bad_pixels

    def __call__(self, array: DiffractionArray) -> DiffractionArray:
        patterns = array.get_patterns()

        if patterns.ndim == 2:
            patterns = patterns[numpy.newaxis, ...]
        elif patterns.ndim != 3:
            raise ValueError(f'Invalid diffraction pattern dimensions! (shape={patterns.shape})')

        if self.filter_values is not None:
            patterns = self.filter_values.apply(patterns)

        if self.crop is not None:
            patterns = self.crop.apply(patterns)

        if self.binning is not None:
            patterns = self.binning.apply(patterns)

        if self.padding is not None:
            patterns = self.padding.apply(patterns)

        if self.hflip:
            patterns = numpy.flip(patterns, axis=-1)

        if self.vflip:
            patterns = numpy.flip(patterns, axis=-2)

        if self.transpose:
            patterns = numpy.transpose(patterns, axes=(0, 2, 1))

        return SimpleDiffractionArray(array.get_label(), array.get_indexes(), patterns)



from pathlib import Path

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class DetectorSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Detector')
        self._group.add_observer(self)

        self.width_px = self._group.create_integer_parameter('WidthInPixels', 1024, minimum=1)
        self.pixel_width_m = self._group.create_real_parameter(
            'PixelWidthInMeters', 75e-6, minimum=0.0
        )
        self.height_px = self._group.create_integer_parameter('HeightInPixels', 1024, minimum=1)
        self.pixel_height_m = self._group.create_real_parameter(
            'PixelHeightInMeters', 75e-6, minimum=0.0
        )
        self.bit_depth = self._group.create_integer_parameter('BitDepth', 8, minimum=1)

        self.bad_pixels_file_type = self._group.create_string_parameter(
            'BadPixelsFileType', 'NPY_Bad_Pixels'
        )
        self.bad_pixels_file_path = self._group.create_path_parameter(
            'BadPixelsFilePath', Path('/path/to/bad_pixels.npy')
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class DiffractionSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Diffraction')
        self._group.add_observer(self)

        self.file_type = self._group.create_string_parameter('FileType', 'NPY')
        self.file_path = self._group.create_path_parameter('FilePath', Path('/path/to/data.npy'))
        self.memmap_enabled = self._group.create_boolean_parameter('MemmapEnabled', False)
        self.scratch_directory = self._group.create_path_parameter(
            'ScratchDirectory', Path.home() / '.ptychodus'
        )

        self.crop_enabled = self._group.create_boolean_parameter('CropEnabled', True)
        self.crop_center_x_px = self._group.create_integer_parameter(
            'CropCenterXInPixels', 32, minimum=0
        )
        self.crop_center_y_px = self._group.create_integer_parameter(
            'CropCenterYInPixels', 32, minimum=0
        )
        self.crop_width_px = self._group.create_integer_parameter(
            'CropWidthInPixels', 64, minimum=1
        )
        self.crop_height_px = self._group.create_integer_parameter(
            'CropHeightInPixels', 64, minimum=1
        )

        self.binning_enabled = self._group.create_boolean_parameter('BinningEnabled', False)
        self.bin_size_x = self._group.create_integer_parameter('BinSizeX', 1, minimum=1)
        self.bin_size_y = self._group.create_integer_parameter('BinSizeY', 1, minimum=1)

        self.padding_enabled = self._group.create_boolean_parameter('PaddingEnabled', False)
        self.pad_x = self._group.create_integer_parameter('PadX', 0, minimum=0)
        self.pad_y = self._group.create_integer_parameter('PadY', 0, minimum=0)

        self.hflip = self._group.create_boolean_parameter('FlipHorizontal', False)
        self.vflip = self._group.create_boolean_parameter('FlipVertical', False)
        self.transpose = self._group.create_boolean_parameter('Transpose', False)

        self.value_lower_bound_enabled = self._group.create_boolean_parameter(
            'ValueLowerBoundEnabled', False
        )
        self.value_lower_bound = self._group.create_integer_parameter(
            'ValueLowerBound', 0, minimum=0
        )
        self.value_upper_bound_enabled = self._group.create_boolean_parameter(
            'ValueUpperBoundEnabled', False
        )
        self.value_upper_bound = self._group.create_integer_parameter(
            'ValueUpperBound', 65535, minimum=0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from __future__ import annotations
import logging

from PyQt5.QtCore import QModelIndex, QStringListModel
from PyQt5.QtWidgets import QAbstractItemView, QDialog

from ptychodus.api.observer import SequenceObserver

from ...model.analysis import IlluminationMapper, ProbePropagator
from ...model.fluorescence import FluorescenceEnhancer
from ...model.product import ProbeAPI, ProbeRepository
from ...model.product.probe import ProbeRepositoryItem
from ...model.visualization import VisualizationEngine
from ...view.repository import RepositoryTreeView
from ...view.widgets import (
    ComboBoxItemDelegate,
    ExceptionDialog,
    ProgressBarItemDelegate,
)
from ..data import FileDialogFactory
from ..helpers import connect_triggered_signal, create_brush_for_editable_cell
from ..image import ImageController
from .editor_factory import ProbeEditorViewControllerFactory
from .fluorescence import FluorescenceViewController
from .illumination import IlluminationViewController
from .propagator import ProbePropagationViewController
from .tree_model import ProbeTreeModel

logger = logging.getLogger(__name__)


class ProbeController(SequenceObserver[ProbeRepositoryItem]):
    def __init__(
        self,
        repository: ProbeRepository,
        api: ProbeAPI,
        image_controller: ImageController,
        propagator: ProbePropagator,
        propagator_visualization_engine: VisualizationEngine,
        illumination_mapper: IlluminationMapper,
        illumination_visualization_engine: VisualizationEngine,
        fluorescence_enhancer: FluorescenceEnhancer,
        fluorescence_visualization_engine: VisualizationEngine,
        view: RepositoryTreeView,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._repository = repository
        self._api = api
        self._image_controller = image_controller
        self._view = view
        self._file_dialog_factory = file_dialog_factory
        self._tree_model = ProbeTreeModel(repository, api, create_brush_for_editable_cell(view))
        self._editor_factory = ProbeEditorViewControllerFactory()

        self._propagation_view_controller = ProbePropagationViewController(
            propagator, propagator_visualization_engine, file_dialog_factory
        )
        self._illumination_view_controller = IlluminationViewController(
            illumination_mapper,
            illumination_visualization_engine,
            file_dialog_factory,
        )
        self._fluorescence_view_controller = FluorescenceViewController(
            fluorescence_enhancer, fluorescence_visualization_engine, file_dialog_factory
        )

        # TODO figure out good fix when saving NPY file without suffix (numpy adds suffix)
        repository.add_observer(self)

        builder_list_model = QStringListModel()
        builder_list_model.setStringList([name for name in api.builder_names()])
        builder_item_delegate = ComboBoxItemDelegate(builder_list_model, view.tree_view)

        view.tree_view.setModel(self._tree_model)
        view.tree_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        power_item_delegate = ProgressBarItemDelegate(view.tree_view)
        view.tree_view.setItemDelegateForColumn(1, power_item_delegate)
        view.tree_view.setItemDelegateForColumn(2, builder_item_delegate)
        selection_model = view.tree_view.selectionModel()

        if selection_model is None:
            raise ValueError('selection_model is None!')
        else:
            selection_model.currentChanged.connect(self._update_view)

        self._update_view(QModelIndex(), QModelIndex())

        load_from_file_action = view.button_box.load_menu.addAction('Open File...')
        connect_triggered_signal(load_from_file_action, self._load_current_probe_from_file)

        copy_action = view.button_box.load_menu.addAction('Copy...')
        connect_triggered_signal(copy_action, self._copy_to_current_probe)

        save_to_file_action = view.button_box.save_menu.addAction('Save File...')
        connect_triggered_signal(save_to_file_action, self._save_current_probe_to_file)

        sync_to_settings_action = view.button_box.save_menu.addAction('Sync To Settings')
        connect_triggered_signal(sync_to_settings_action, self._sync_current_probe_to_settings)

        view.copier_dialog.setWindowTitle('Copy Probe')
        view.copier_dialog.source_combo_box.setModel(self._tree_model)
        view.copier_dialog.destination_combo_box.setModel(self._tree_model)
        view.copier_dialog.finished.connect(self._finish_copying_probe)

        view.button_box.edit_button.clicked.connect(self._edit_current_probe)

        propagate_action = view.button_box.analyze_menu.addAction('Propagate...')
        connect_triggered_signal(propagate_action, self._propagate_probe)

        illumination_action = view.button_box.analyze_menu.addAction('Map Illumination...')
        connect_triggered_signal(illumination_action, self._map_illumination)

        fluorescence_action = view.button_box.analyze_menu.addAction('Enhance Fluorescence...')
        connect_triggered_signal(fluorescence_action, self._enhance_fluorescence)

    def _get_current_item_index(self) -> int:
        model_index = self._view.tree_view.currentIndex()

        if model_index.isValid():
            parent = model_index.parent()

            while parent.isValid():
                model_index = parent
                parent = model_index.parent()

            return model_index.row()

        logger.warning('No current index!')
        return -1

    def _load_current_probe_from_file(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        file_path, name_filter = self._file_dialog_factory.get_open_file_path(
            self._view,
            'Open Probe',
            name_filters=[nf for nf in self._api.get_open_file_filters()],
            selected_name_filter=self._api.get_open_file_filter(),
        )

        if file_path:
            try:
                self._api.open_probe(item_index, file_path, file_type=name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception('File Reader', err)

    def _copy_to_current_probe(self) -> None:
        item_index = self._get_current_item_index()

        if item_index >= 0:
            self._view.copier_dialog.destination_combo_box.setCurrentIndex(item_index)
            self._view.copier_dialog.open()

    def _finish_copying_probe(self, result: int) -> None:
        if result == QDialog.DialogCode.Accepted:
            source_index = self._view.copier_dialog.source_combo_box.currentIndex()
            destination_index = self._view.copier_dialog.destination_combo_box.currentIndex()
            self._api.copy_probe(source_index, destination_index)

    def _edit_current_probe(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        item_name = self._repository.get_name(item_index)
        item = self._repository[item_index]
        dialog = self._editor_factory.create_editor_dialog(item_name, item, self._view)
        dialog.open()

    def _save_current_probe_to_file(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        file_path, name_filter = self._file_dialog_factory.get_save_file_path(
            self._view,
            'Save Probe',
            name_filters=[nf for nf in self._api.get_save_file_filters()],
            selected_name_filter=self._api.get_save_file_filter(),
        )

        if file_path:
            try:
                self._api.save_probe(item_index, file_path, name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception('File Writer', err)

    def _sync_current_probe_to_settings(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            item = self._repository[item_index]
            item.sync_to_settings()

    def _propagate_probe(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            self._propagation_view_controller.launch(item_index)

    def _map_illumination(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            self._illumination_view_controller.map(item_index)

    def _enhance_fluorescence(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            self._fluorescence_view_controller.launch(item_index)

    def _update_view(self, current: QModelIndex, previous: QModelIndex) -> None:
        enabled = current.isValid()
        self._view.button_box.load_button.setEnabled(enabled)
        self._view.button_box.save_button.setEnabled(enabled)
        self._view.button_box.edit_button.setEnabled(enabled)
        self._view.button_box.analyze_button.setEnabled(enabled)

        item_index = self._get_current_item_index()

        if item_index < 0:
            self._image_controller.clear_array()
        else:
            try:
                item = self._repository[item_index]
            except IndexError:
                logger.warning('Unable to access item for visualization!')
            else:
                probe = item.get_probes().get_probe_no_opr()  # TODO OPR
                array = (
                    probe.get_incoherent_mode(current.row())
                    if current.parent().isValid()
                    else probe.get_incoherent_modes_flattened()
                )
                self._image_controller.set_array(array, probe.get_pixel_geometry())

    def handle_item_inserted(self, index: int, item: ProbeRepositoryItem) -> None:
        self._tree_model.insert_item(index, item)

    def handle_item_changed(self, index: int, item: ProbeRepositoryItem) -> None:
        self._tree_model.update_item(index, item)

        if index == self._get_current_item_index():
            current_index = self._view.tree_view.currentIndex()
            self._update_view(current_index, current_index)

    def handle_item_removed(self, index: int, item: ProbeRepositoryItem) -> None:
        self._tree_model.remove_item(index, item)



from __future__ import annotations
from pathlib import Path
from uuid import UUID

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class GlobusSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Globus')
        self._group.add_observer(self)

        # Transfer copies only those files where the source file differs from
        # the destination file. The sync level is an integer in the 0-3 range,
        # and it controls what checks are used to determine if a file is
        # different. Higher levels include the checks from lower levels. Delete
        # tasks are always null:
        #
        # 0 Copy files that do not exist at the destination.
        # 1 Copy files if the size of the destination does not match the size of the source.
        # 2 Copy files if the timestamp of the destination is older than the timestamp of the source.
        # 3 Copy files if checksums of the source and destination do not match.
        self.transfer_sync_level = self._group.create_integer_parameter(
            'TransferSyncLevel', 3, minimum=0, maximum=3
        )

        self.input_data_endpoint_id = self._group.create_uuid_parameter(
            'InputDataEndpointID', UUID(int=0)
        )
        self.input_data_globus_path = self._group.create_string_parameter(
            'InputDataGlobusPath', '/~/path/to/input/data'
        )
        self.input_data_posix_path = self._group.create_path_parameter(
            'InputDataPosixPath', Path('/path/to/input/data')
        )

        self.compute_endpoint_id = self._group.create_uuid_parameter(
            'ComputeEndpointID', UUID(int=0)
        )
        self.compute_data_endpoint_id = self._group.create_uuid_parameter(
            'ComputeDataEndpointID', UUID(int=0)
        )
        self.compute_data_globus_path = self._group.create_string_parameter(
            'ComputeDataGlobusPath', '/~/path/to/compute/data'
        )
        self.compute_data_posix_path = self._group.create_path_parameter(
            'ComputeDataPosixPath', Path('/path/to/compute/data')
        )

        self.status_auto_refresh = self._group.create_boolean_parameter('StatusAutoRefresh', False)
        self.status_refresh_interval_s = self._group.create_integer_parameter(
            'StatusRefreshIntervalInSeconds', 30, minimum=10, maximum=86400
        )

        self.output_data_endpoint_id = self._group.create_uuid_parameter(
            'OutputDataEndpointID', UUID(int=0)
        )
        self.output_data_globus_path = self._group.create_string_parameter(
            'OutputDataGlobusPath', '/~/path/to/output/data'
        )
        self.output_data_posix_path = self._group.create_path_parameter(
            'OutputDataPosixPath', Path('/path/to/output/data')
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from __future__ import annotations
from collections.abc import Sequence
from pathlib import Path
from typing import Any
import logging

import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.observer import Observable
from ptychodus.api.probe import ProbeSequence
from ptychodus.api.propagator import (
    AngularSpectrumPropagator,
    PropagatorParameters,
    ComplexArrayType,
    intensity,
)

from ..product import ProductRepository
from .settings import ProbePropagationSettings

logger = logging.getLogger(__name__)


class ProbePropagator(Observable):
    def __init__(self, settings: ProbePropagationSettings, repository: ProductRepository) -> None:
        super().__init__()
        self._settings = settings
        self._repository = repository

        self._product_index = -1
        self._propagated_wavefield: ComplexArrayType | None = None
        self._propagated_intensity: RealArrayType | None = None

    def set_product(self, product_index: int) -> None:
        if self._product_index != product_index:
            self._product_index = product_index
            self._propagated_wavefield = None
            self._propagated_intensity = None
            self.notify_observers()

    def get_product_name(self) -> str:
        item = self._repository[self._product_index]
        return item.get_name()

    def propagate(
        self,
        *,
        begin_coordinate_m: float,
        end_coordinate_m: float,
        num_steps: int,
    ) -> None:
        item = self._repository[self._product_index]
        probe = item.get_probe_item().get_probes().get_probe_no_opr()  # TODO OPR
        wavelength_m = item.get_geometry().probe_wavelength_m
        propagated_wavefield = numpy.zeros(
            (num_steps, probe.num_incoherent_modes, probe.height_px, probe.width_px),
            dtype=probe.dtype,
        )
        propagated_intensity = numpy.zeros((num_steps, probe.height_px, probe.width_px))
        distance_m = numpy.linspace(begin_coordinate_m, end_coordinate_m, num_steps)
        pixel_geometry = probe.get_pixel_geometry()

        for idx, z_m in enumerate(distance_m):
            propagator_parameters = PropagatorParameters(
                wavelength_m=wavelength_m,
                width_px=probe.width_px,
                height_px=probe.height_px,
                pixel_width_m=pixel_geometry.width_m,
                pixel_height_m=pixel_geometry.height_m,
                propagation_distance_m=float(z_m),
            )
            propagator = AngularSpectrumPropagator(propagator_parameters)

            for mode in range(probe.num_incoherent_modes):
                wf = propagator.propagate(probe.get_incoherent_mode(mode))
                propagated_wavefield[idx, mode, :, :] = wf
                propagated_intensity[idx, :, :] += intensity(wf)

        self._settings.begin_coordinate_m.set_value(begin_coordinate_m)
        self._settings.end_coordinate_m.set_value(end_coordinate_m)
        self._propagated_wavefield = propagated_wavefield
        self._propagated_intensity = propagated_intensity
        self.notify_observers()

    def get_begin_coordinate_m(self) -> float:
        return self._settings.begin_coordinate_m.get_value()

    def get_end_coordinate_m(self) -> float:
        return self._settings.end_coordinate_m.get_value()

    def _get_probe(self) -> ProbeSequence:
        item = self._repository[self._product_index]
        return item.get_probe_item().get_probes()

    def get_pixel_geometry(self) -> PixelGeometry | None:
        try:
            probe = self._get_probe()
        except IndexError:
            return None
        else:
            return probe.get_pixel_geometry()

    def get_num_steps(self) -> int:
        if self._propagated_intensity is None:
            return self._settings.num_steps.get_value()

        return self._propagated_intensity.shape[0]

    def get_xy_projection(self, step: int) -> RealArrayType:
        if self._propagated_intensity is None:
            raise ValueError('No propagated wavefield!')

        return self._propagated_intensity[step]

    def get_zx_projection(self) -> RealArrayType:
        if self._propagated_intensity is None:
            raise ValueError('No propagated wavefield!')

        sz = self._propagated_intensity.shape[-2]
        cut_plane_l = self._propagated_intensity[:, (sz - 1) // 2, :]
        cut_plane_r = self._propagated_intensity[:, sz // 2, :]
        return numpy.transpose(numpy.add(cut_plane_l, cut_plane_r) / 2)

    def get_zy_projection(self) -> RealArrayType:
        if self._propagated_intensity is None:
            raise ValueError('No propagated wavefield!')

        sz = self._propagated_intensity.shape[-1]
        cut_plane_l = self._propagated_intensity[:, :, (sz - 1) // 2]
        cut_plane_r = self._propagated_intensity[:, :, sz // 2]
        return numpy.transpose(numpy.add(cut_plane_l, cut_plane_r) / 2)

    def get_save_file_filters(self) -> Sequence[str]:
        return [self.get_save_file_filter()]

    def get_save_file_filter(self) -> str:
        return 'NumPy Zipped Archive (*.npz)'

    def save_propagated_probe(self, file_path: Path) -> None:
        if self._propagated_wavefield is None or self._propagated_intensity is None:
            raise ValueError('No propagated wavefield!')

        contents: dict[str, Any] = {
            'begin_coordinate_m': self.get_begin_coordinate_m(),
            'end_coordinate_m': self.get_end_coordinate_m(),
            'wavefield': self._propagated_wavefield,
            'intensity': self._propagated_intensity,
        }

        pixel_geometry = self.get_pixel_geometry()

        if pixel_geometry is not None:
            contents['pixel_height_m'] = pixel_geometry.height_m
            contents['pixel_width_m'] = pixel_geometry.width_m

        numpy.savez_compressed(file_path, allow_pickle=False, **contents)



from typing import Any
import logging

from PyQt5.QtCore import QAbstractListModel, QModelIndex, QObject, Qt

from ...model.product import ProductRepository

logger = logging.getLogger(__name__)


class ProductRepositoryListModel(QAbstractListModel):
    def __init__(
        self,
        repository: ProductRepository,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)
        self._repository = repository

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid() and role == Qt.ItemDataRole.DisplayRole:
            try:
                item = self._repository[index.row()]
            except IndexError as err:
                logger.exception(err)
            else:
                return item.get_name()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._repository)



import logging

import numpy.random

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider

from .builder import FromMemoryProbeBuilder
from .builder_factory import ProbeBuilderFactory
from .item import ProbeRepositoryItem
from .multimodal import MultimodalProbeBuilder
from .settings import ProbeSettings

logger = logging.getLogger(__name__)


class ProbeRepositoryItemFactory:
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ProbeSettings,
        builder_factory: ProbeBuilderFactory,
    ) -> None:
        self._rng = rng
        self._settings = settings
        self._builder_factory = builder_factory

    def create(
        self, geometry_provider: ProbeGeometryProvider, probe: ProbeSequence | None = None
    ) -> ProbeRepositoryItem:
        multimodal_builder = MultimodalProbeBuilder(self._rng, self._settings)

        if probe is None:
            builder = self._builder_factory.create_default()
        else:
            builder = FromMemoryProbeBuilder(self._settings, probe)
            multimodal_builder.set_identity()

        return ProbeRepositoryItem(geometry_provider, self._settings, builder, multimodal_builder)

    def create_from_settings(self, geometry_provider: ProbeGeometryProvider) -> ProbeRepositoryItem:
        try:
            builder = self._builder_factory.create_from_settings()
        except Exception as exc:
            logger.error(''.join(exc.args))
            builder = self._builder_factory.create_default()

        multimodal_builder = MultimodalProbeBuilder(self._rng, self._settings)
        return ProbeRepositoryItem(geometry_provider, self._settings, builder, multimodal_builder)



from pathlib import Path
from typing import Final
import logging

import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionFileWriter,
    DiffractionMetadata,
    SimpleDiffractionDataset,
    SimpleDiffractionArray,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.tree import SimpleTreeNode

logger = logging.getLogger(__name__)


class NPZDiffractionFileIO(DiffractionFileReader, DiffractionFileWriter):
    SIMPLE_NAME: Final[str] = 'NPZ'
    DISPLAY_NAME: Final[str] = 'Ptychodus NumPy Zipped Archive (*.npz)'

    INDEXES: Final[str] = 'indexes'
    PATTERNS: Final[str] = 'patterns'
    BAD_PIXELS: Final[str] = 'bad_pixels'

    def read(self, file_path: Path) -> DiffractionDataset:
        dataset = SimpleDiffractionDataset.create_null(file_path)
        contents = numpy.load(file_path)

        try:
            patterns = contents[self.PATTERNS]
        except KeyError:
            logger.warning(f'Failed to read patterns in "{file_path}".')
            return dataset

        num_patterns, detector_height, detector_width = patterns.shape

        metadata = DiffractionMetadata(
            num_patterns_per_array=[num_patterns],
            pattern_dtype=patterns.dtype,
            detector_extent=ImageExtent(detector_width, detector_height),
            file_path=file_path,
        )

        try:
            indexes = contents[self.INDEXES]
        except KeyError:
            logger.warning(f'Failed to read indexes in "{file_path}".')
            indexes = numpy.arange(num_patterns)

        contents_tree = SimpleTreeNode.create_root(['Name', 'Type', 'Details'])
        contents_tree.create_child(
            [
                file_path.stem,
                type(patterns).__name__,
                f'{patterns.dtype}{patterns.shape}',
            ]
        )

        array = SimpleDiffractionArray(
            label=file_path.stem,
            indexes=indexes,
            patterns=patterns,
        )

        return SimpleDiffractionDataset(
            metadata, contents_tree, [array], bad_pixels=contents.get(self.BAD_PIXELS, None)
        )

    def write(self, file_path: Path, dataset: DiffractionDataset) -> None:
        contents = {
            self.INDEXES: numpy.concatenate([array.get_indexes() for array in dataset]),
            self.PATTERNS: numpy.concatenate([array.get_patterns() for array in dataset]),
        }

        bad_pixels = dataset.get_bad_pixels()

        if bad_pixels is not None:
            contents[self.BAD_PIXELS] = bad_pixels

        numpy.savez_compressed(file_path, allow_pickle=False, **contents)


def register_plugins(registry: PluginRegistry) -> None:
    npz_diffraction_file_io = NPZDiffractionFileIO()

    registry.diffraction_file_readers.register_plugin(
        npz_diffraction_file_io,
        simple_name=NPZDiffractionFileIO.SIMPLE_NAME,
        display_name=NPZDiffractionFileIO.DISPLAY_NAME,
    )
    registry.diffraction_file_writers.register_plugin(
        npz_diffraction_file_io,
        simple_name=NPZDiffractionFileIO.SIMPLE_NAME,
        display_name=NPZDiffractionFileIO.DISPLAY_NAME,
    )



from __future__ import annotations
import logging

from PyQt5.QtCore import QModelIndex, QStringListModel
from PyQt5.QtWidgets import QAbstractItemView, QDialog

from ptychodus.api.observer import SequenceObserver

from ...model.analysis import FourierAnalyzer, FourierRingCorrelator, XMCDAnalyzer
from ...model.product import ObjectAPI, ObjectRepository
from ...model.product.object import ObjectRepositoryItem
from ...model.visualization import VisualizationEngine
from ...view.repository import RepositoryTreeView
from ...view.widgets import ComboBoxItemDelegate, ExceptionDialog
from ..data import FileDialogFactory
from ..helpers import connect_triggered_signal, create_brush_for_editable_cell
from ..image import ImageController
from .editor_factory import ObjectEditorViewControllerFactory
from .fourier import FourierAnalysisViewController
from .frc import FourierRingCorrelationViewController
from .tree_model import ObjectTreeModel
from .xmcd import XMCDViewController

logger = logging.getLogger(__name__)


class ObjectController(SequenceObserver[ObjectRepositoryItem]):
    def __init__(
        self,
        repository: ObjectRepository,
        api: ObjectAPI,
        image_controller: ImageController,
        correlator: FourierRingCorrelator,
        fourier_analyzer: FourierAnalyzer,
        fourier_real_space_visualization_engine: VisualizationEngine,
        fourier_reciprocal_space_visualization_engine: VisualizationEngine,
        xmcd_analyzer: XMCDAnalyzer,
        xmcd_visualization_engine: VisualizationEngine,
        view: RepositoryTreeView,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._repository = repository
        self._api = api
        self._image_controller = image_controller
        self._view = view
        self._file_dialog_factory = file_dialog_factory
        self._tree_model = ObjectTreeModel(repository, api, create_brush_for_editable_cell(view))
        self._editor_factory = ObjectEditorViewControllerFactory()

        self._frc_view_controller = FourierRingCorrelationViewController(
            correlator, self._tree_model
        )
        self._fourier_view_controller = FourierAnalysisViewController(
            fourier_analyzer,
            fourier_real_space_visualization_engine,
            fourier_reciprocal_space_visualization_engine,
            file_dialog_factory,
        )
        self._xmcd_view_controller = XMCDViewController(
            xmcd_analyzer, xmcd_visualization_engine, file_dialog_factory, self._tree_model
        )

        # TODO figure out good fix when saving NPY file without suffix (numpy adds suffix)
        repository.add_observer(self)

        builder_list_model = QStringListModel()
        builder_list_model.setStringList([name for name in api.builder_names()])
        builder_item_delegate = ComboBoxItemDelegate(builder_list_model, view.tree_view)

        view.tree_view.setModel(self._tree_model)
        view.tree_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        view.tree_view.setItemDelegateForColumn(2, builder_item_delegate)
        selection_model = view.tree_view.selectionModel()

        if selection_model is None:
            raise ValueError('selection_model is None!')
        else:
            selection_model.currentChanged.connect(self._update_view)

        self._update_view(QModelIndex(), QModelIndex())

        load_from_file_action = view.button_box.load_menu.addAction('Open File...')
        connect_triggered_signal(load_from_file_action, self._load_current_object_from_file)

        copy_action = view.button_box.load_menu.addAction('Copy...')
        connect_triggered_signal(copy_action, self._copy_to_current_object)

        save_to_file_action = view.button_box.save_menu.addAction('Save File...')
        connect_triggered_signal(save_to_file_action, self._save_current_object_to_file)

        sync_to_settings_action = view.button_box.save_menu.addAction('Sync To Settings')
        connect_triggered_signal(sync_to_settings_action, self._sync_current_object_to_settings)

        view.copier_dialog.setWindowTitle('Copy Object')
        view.copier_dialog.source_combo_box.setModel(self._tree_model)
        view.copier_dialog.destination_combo_box.setModel(self._tree_model)
        view.copier_dialog.finished.connect(self._finish_copying_object)

        view.button_box.edit_button.clicked.connect(self._edit_current_object)

        frc_action = view.button_box.analyze_menu.addAction('Fourier Ring Correlation...')
        connect_triggered_signal(frc_action, self._analyze_frc)

        fourier_action = view.button_box.analyze_menu.addAction('Fourier Analysis...')
        connect_triggered_signal(fourier_action, self._analyze_fourier)

        xmcd_action = view.button_box.analyze_menu.addAction('XMCD...')
        connect_triggered_signal(xmcd_action, self._analyze_xmcd)

    def _get_current_item_index(self) -> int:
        model_index = self._view.tree_view.currentIndex()

        if model_index.isValid():
            parent = model_index.parent()

            while parent.isValid():
                model_index = parent
                parent = model_index.parent()

            return model_index.row()

        logger.warning('No current index!')
        return -1

    def _load_current_object_from_file(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        file_path, name_filter = self._file_dialog_factory.get_open_file_path(
            self._view,
            'Open Object',
            name_filters=[nf for nf in self._api.get_open_file_filters()],
            selected_name_filter=self._api.get_open_file_filter(),
        )

        if file_path:
            try:
                self._api.open_object(item_index, file_path, file_type=name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception('File Reader', err)

    def _copy_to_current_object(self) -> None:
        item_index = self._get_current_item_index()

        if item_index >= 0:
            self._view.copier_dialog.destination_combo_box.setCurrentIndex(item_index)
            self._view.copier_dialog.open()

    def _finish_copying_object(self, result: int) -> None:
        if result == QDialog.DialogCode.Accepted:
            source_index = self._view.copier_dialog.source_combo_box.currentIndex()
            destination_index = self._view.copier_dialog.destination_combo_box.currentIndex()
            self._api.copy_object(source_index, destination_index)

    def _edit_current_object(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        item_name = self._repository.get_name(item_index)
        item = self._repository[item_index]
        dialog = self._editor_factory.create_editor_dialog(item_name, item, self._view)
        dialog.open()

    def _save_current_object_to_file(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        file_path, name_filter = self._file_dialog_factory.get_save_file_path(
            self._view,
            'Save Object',
            name_filters=[nameFilter for nameFilter in self._api.get_save_file_filters()],
            selected_name_filter=self._api.get_save_file_filter(),
        )

        if file_path:
            try:
                self._api.save_object(item_index, file_path, name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception('File Writer', err)

    def _sync_current_object_to_settings(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            item = self._repository[item_index]
            item.sync_to_settings()

    def _analyze_frc(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            self._frc_view_controller.analyze(item_index, item_index)

    def _analyze_fourier(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            self._fourier_view_controller.analyze(item_index)

    def _analyze_xmcd(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            self._xmcd_view_controller.analyze(item_index, item_index)

    def _update_view(self, current: QModelIndex, previous: QModelIndex) -> None:
        enabled = current.isValid()
        self._view.button_box.load_button.setEnabled(enabled)
        self._view.button_box.save_button.setEnabled(enabled)
        self._view.button_box.edit_button.setEnabled(enabled)
        self._view.button_box.analyze_button.setEnabled(enabled)

        item_index = self._get_current_item_index()

        if item_index < 0:
            self._image_controller.clear_array()
        else:
            try:
                item = self._repository[item_index]
            except IndexError:
                logger.warning('Unable to access item for visualization!')
            else:
                object_ = item.get_object()
                array = (
                    object_.get_layer(current.row())
                    if current.parent().isValid()
                    else object_.get_layers_flattened()
                )
                self._image_controller.set_array(array, object_.get_pixel_geometry())

    def handle_item_inserted(self, index: int, item: ObjectRepositoryItem) -> None:
        self._tree_model.insert_item(index, item)

    def handle_item_changed(self, index: int, item: ObjectRepositoryItem) -> None:
        self._tree_model.update_item(index, item)

        if index == self._get_current_item_index():
            current_index = self._view.tree_view.currentIndex()
            self._update_view(current_index, current_index)

    def handle_item_removed(self, index: int, item: ObjectRepositoryItem) -> None:
        self._tree_model.remove_item(index, item)



from __future__ import annotations
from collections.abc import Iterator

from matplotlib.colors import Normalize
import numpy

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.visualization import (
    NumberArrayType,
    RealArrayType,
    VisualizationProduct,
)

from .color_axis import ColorAxis
from .color_model import CylindricalColorModelParameter
from .components import AmplitudeArrayComponent, PhaseInRadiansArrayComponent
from .renderer import Renderer
from .transformation import ScalarTransformationParameter


class CylindricalColorModelRenderer(Renderer):
    def __init__(
        self,
        amplitude_component: AmplitudeArrayComponent,
        phase_component: PhaseInRadiansArrayComponent,
        transformation: ScalarTransformationParameter,
        color_axis: ColorAxis,
    ) -> None:
        super().__init__('Complex')
        self._amplitude_component = amplitude_component
        self._phase_component = phase_component
        self._transformation = transformation
        self._add_parameter('transformation', transformation)
        self._color_axis = color_axis
        self._add_group('color_axis', color_axis, observe=True)
        self._color_model = CylindricalColorModelParameter()
        self._add_parameter('color_model', self._color_model)

    def variants(self) -> Iterator[str]:
        return self._color_model.choices()

    def get_variant(self) -> str:
        return self._color_model.get_value()

    def set_variant(self, variant: str) -> None:
        self._color_model.set_value(variant)

    def is_cyclic(self) -> bool:
        return True

    def _colorize(self, amplitude: RealArrayType, phase_rad: RealArrayType) -> RealArrayType:
        vrange = self._color_axis.get_range()
        norm = Normalize(vmin=vrange.lower, vmax=vrange.upper, clip=False)

        model = self._color_model.get_plugin()
        h = (phase_rad + numpy.pi) / (2 * numpy.pi)
        return model(h, norm(amplitude))

    def colorize(self, array: NumberArrayType) -> RealArrayType:
        amplitude = self._amplitude_component.calculate(array)
        amplitude_transformed = self._transformation.transform(amplitude)
        phase_rad = self._phase_component.calculate(array)
        return self._colorize(amplitude_transformed, phase_rad)

    def render(
        self, array: NumberArrayType, pixel_geometry: PixelGeometry, *, autoscale_color_axis: bool
    ) -> VisualizationProduct:
        amplitude = self._amplitude_component.calculate(array)
        amplitude_transformed = self._transformation.transform(amplitude)
        phase_rad = self._phase_component.calculate(array)

        if autoscale_color_axis:
            self._color_axis.set_to_data_range(amplitude_transformed)

        rgba = self._colorize(amplitude_transformed, phase_rad)

        return VisualizationProduct(
            value_label=self._transformation.decorate_text(self._amplitude_component.name),
            values=array,
            rgba=rgba,
            pixel_geometry=pixel_geometry,
        )



from collections.abc import Callable, Iterable, Iterator, Mapping
from pathlib import Path
import logging

import numpy

from ptychodus.api.object import Object, ObjectFileReader, ObjectFileWriter
from ptychodus.api.plugins import PluginChooser

from ...diffraction import DiffractionAPI
from .builder import FromFileObjectBuilder, ObjectBuilder
from .dead_leaves import DeadLeavesObjectBuilder
from .random import RandomObjectBuilder
from .settings import ObjectSettings
from .stxm import STXMObjectBuilder

logger = logging.getLogger(__name__)


class ObjectBuilderFactory(Iterable[str]):
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ObjectSettings,
        diffraction_api: DiffractionAPI,
        file_reader_chooser: PluginChooser[ObjectFileReader],
        file_writer_chooser: PluginChooser[ObjectFileWriter],
    ) -> None:
        self._settings = settings
        self._file_reader_chooser = file_reader_chooser
        self._file_writer_chooser = file_writer_chooser
        self._builders: Mapping[str, Callable[[], ObjectBuilder]] = {
            'random': lambda: RandomObjectBuilder(rng, settings),
            'dead_leaves': lambda: DeadLeavesObjectBuilder(rng, settings),
            'stxm': lambda: STXMObjectBuilder(settings, diffraction_api),
        }

    def __iter__(self) -> Iterator[str]:
        return iter(self._builders)

    def create(self, name: str) -> ObjectBuilder:
        try:
            factory = self._builders[name]
        except KeyError as exc:
            raise KeyError(f'Unknown object builder "{name}"!') from exc

        return factory()

    def create_default(self) -> ObjectBuilder:
        return next(iter(self._builders.values()))()

    def create_from_settings(self) -> ObjectBuilder:
        name = self._settings.builder.get_value()
        name_repaired = name.casefold()

        if name_repaired == 'from_file':
            return self.create_object_from_file(
                self._settings.file_path.get_value(),
                self._settings.file_type.get_value(),
            )

        return self.create(name_repaired)

    def get_open_file_filters(self) -> Iterator[str]:
        for plugin in self._file_reader_chooser:
            yield plugin.display_name

    def get_open_file_filter(self) -> str:
        return self._file_reader_chooser.get_current_plugin().display_name

    def create_object_from_file(self, file_path: Path, file_filter: str) -> ObjectBuilder:
        self._file_reader_chooser.set_current_plugin(file_filter)
        file_type = self._file_reader_chooser.get_current_plugin().simple_name
        file_reader = self._file_reader_chooser.get_current_plugin().strategy
        return FromFileObjectBuilder(self._settings, file_path, file_type, file_reader)

    def get_save_file_filters(self) -> Iterator[str]:
        for plugin in self._file_writer_chooser:
            yield plugin.display_name

    def get_save_file_filter(self) -> str:
        return self._file_writer_chooser.get_current_plugin().display_name

    def save_object(self, file_path: Path, file_filter: str, object_: Object) -> None:
        self._file_writer_chooser.set_current_plugin(file_filter)
        file_type = self._file_writer_chooser.get_current_plugin().simple_name
        logger.debug(f'Writing "{file_path}" as "{file_type}"')
        file_writer = self._file_writer_chooser.get_current_plugin().strategy
        file_writer.write(file_path, object_)



import logging

from ptychodus.api.settings import SettingsRegistry

from ..diffraction import DiffractionAPI
from ..product import ProductAPI
from ..processing import ProcessingAPI
from ..task_manager import TaskManager
from .authorizer import GlobusAuthorizer
from .client import FakeGlobusClient, GlobusClient
from .executor import GlobusExecutor
from .settings import GlobusSettings
from .status import GlobusStatusRepository

logger = logging.getLogger(__name__)


class GlobusCore:
    def __init__(
        self,
        task_manager: TaskManager,
        settings_registry: SettingsRegistry,
        diffraction_api: DiffractionAPI,
        product_api: ProductAPI,
        processing_api: ProcessingAPI,
    ) -> None:
        self.settings = GlobusSettings(settings_registry)
        self.authorizer = GlobusAuthorizer()
        self.status_repository = GlobusStatusRepository()

        try:
            from ._globus_client import RealGlobusClient
        except ModuleNotFoundError:
            logger.info('Globus not found.')
            self._client: GlobusClient = FakeGlobusClient()
        else:
            self._client = RealGlobusClient(
                task_manager, self.settings, self.authorizer, self.status_repository
            )

        self.executor = GlobusExecutor(
            self.settings,
            settings_registry,
            diffraction_api,
            product_api,
            processing_api,
            self._client,
        )

    @property
    def is_supported(self) -> bool:
        return self._client.is_supported

    def start(self) -> None:
        self._client.start()

    def stop(self) -> None:
        self._client.stop()



from __future__ import annotations
from dataclasses import dataclass
import logging

import numpy
import scipy.special

from ptychodus.api.common import RealArrayType
from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ZernikePolynomial:
    radial_degree: int  # n
    angular_frequency: int  # m

    @property
    def spatial_frequencey(self) -> int:
        return self.radial_degree + abs(self.angular_frequency)

    def _radial_polynomial(self, distance: RealArrayType) -> RealArrayType:
        n_minus_m = self.radial_degree - abs(self.angular_frequency)
        half_n_minus_m = n_minus_m // 2
        sgn = 1

        values = numpy.zeros_like(distance)

        for k in range(half_n_minus_m + 1):
            n_minus_k = self.radial_degree - k
            n_minus_2k = self.radial_degree - 2 * k

            coef = sgn
            coef *= scipy.special.binom(n_minus_k, k)
            coef *= scipy.special.binom(n_minus_2k, half_n_minus_m - k)
            coef = int(coef)  # NOTE!

            values += numpy.multiply(coef, numpy.power(distance, n_minus_2k))

            sgn = -sgn

        return values

    def _angular_function(self, angle: RealArrayType) -> RealArrayType:
        return (
            numpy.sin(-self.angular_frequency * angle)
            if self.angular_frequency < 0
            else numpy.cos(self.angular_frequency * angle)
        )

    def __call__(
        self, distance: RealArrayType, angle: RealArrayType, undefined_value: float = 0.0
    ) -> RealArrayType:
        rvalue = self._radial_polynomial(distance)
        avalue = self._angular_function(angle)
        nvalue_sq = self.radial_degree + 1

        if self.angular_frequency != 0:
            nvalue_sq *= 2

        return numpy.where(
            numpy.logical_and(0 < distance, distance <= 1),
            numpy.sqrt(nvalue_sq) * rvalue * avalue,
            undefined_value,
        )

    def __str__(self) -> str:
        return f'$Z_{{{self.radial_degree}}}^{{{self.angular_frequency:+d}}}$'


class ZernikeProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, settings: ProbeSettings) -> None:
        super().__init__(settings, 'zernike')
        self._settings = settings
        self._polynomials: list[ZernikePolynomial] = list()
        self._order = 0

        self.diameter_m = settings.disk_diameter_m.copy()
        self._add_parameter('diameter_m', self.diameter_m)

        # TODO init zernike coefficients from settings
        self.coefficients = self.create_complex_sequence_parameter('coefficients', [1 + 0j])

        self.set_order(1)

    def copy(self) -> ZernikeProbeBuilder:
        builder = ZernikeProbeBuilder(self._settings)
        builder.diameter_m.set_value(self.diameter_m.get_value())
        builder.coefficients.set_value(self.coefficients.get_value())
        builder.set_order(self.get_order())
        return builder

    def set_order(self, order: int) -> None:
        if order < 1:
            logger.warning('Order must be strictly positive!')
            return

        if self._order == order:
            return

        self._polynomials.clear()

        for radial_degree in range(order):
            for angular_frequency in range(-radial_degree, 1 + radial_degree, 2):
                poly = ZernikePolynomial(radial_degree, angular_frequency)
                self._polynomials.append(poly)

        npoly = len(self._polynomials)
        ncoef = len(self.coefficients)

        if ncoef < npoly:
            coef = list(self.coefficients.get_value())
            coef += [0j] * (npoly - ncoef)
            self.coefficients.set_value(coef)

        self._order = order
        self.notify_observers()

    def get_order(self) -> int:
        return self._order

    def set_coefficient(self, idx: int, value: complex) -> None:
        self.coefficients[idx] = value

    def get_coefficient(self, idx: int) -> complex:
        return self.coefficients[idx]

    def get_polynomial(self, idx: int) -> ZernikePolynomial:
        return self._polynomials[idx]

    def __len__(self) -> int:
        return min(len(self.coefficients), len(self._polynomials))

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        geometry = geometry_provider.get_probe_geometry()
        coords = self.get_transverse_coordinates(geometry)

        radius = self.diameter_m.get_value() / 2.0
        distance = numpy.hypot(coords.position_y_m, coords.position_x_m) / radius
        angle = numpy.arctan2(coords.position_y_m, coords.position_x_m)
        array = numpy.zeros_like(distance, dtype=complex)

        for coef, poly in zip(self.coefficients, self._polynomials):
            array += numpy.multiply(coef, poly(distance, angle))

        return ProbeSequence(
            array=self.normalize(array),
            opr_weights=None,
            pixel_geometry=geometry.get_pixel_geometry(),
        )



from __future__ import annotations
from dataclasses import dataclass
from typing import Final

from PyQt5.QtCore import QModelIndex, QPointF, QRectF, QSize, QSizeF, Qt
from PyQt5.QtGui import QBrush, QFontMetrics, QPainter, QPen, QTextDocument, QTextOption
from PyQt5.QtWidgets import QApplication, QStyle, QStyleOptionViewItem, QStyledItemDelegate


@dataclass(frozen=True)
class BubbleMetrics:
    margin_px: int
    border_px: int
    padding_px: int
    radius_px: int

    @property
    def mbp_px(self) -> int:
        return self.margin_px + self.border_px + self.padding_px

    @property
    def bp_px(self) -> int:
        return self.margin_px + self.border_px

    @classmethod
    def from_document(
        cls,
        document: QTextDocument,
        *,
        margin_em: int = 1,
        border_px: int = 1,
        padding_em: int = 1,
        radius_px: int = 10,
    ) -> BubbleMetrics:
        font_metrics = QFontMetrics(document.defaultFont())
        one_em_px = font_metrics.horizontalAdvance('m')
        return cls(
            margin_px=margin_em * one_em_px,
            border_px=border_px,
            padding_px=padding_em * one_em_px,
            radius_px=radius_px,
        )


class ChatBubbleItemDelegate(QStyledItemDelegate):
    TEXT_FRACTIONAL_WIDTH: Final[float] = 0.8

    def _create_text_document(
        self, option: QStyleOptionViewItem, index: QModelIndex
    ) -> QTextDocument:
        text = index.data(Qt.ItemDataRole.DisplayRole)

        text_option = QTextOption()
        text_option.setWrapMode(QTextOption.WrapMode.WordWrap)
        text_option.setTextDirection(option.direction)

        doc = QTextDocument()
        doc.setDefaultTextOption(text_option)
        doc.setHtml(text)
        doc.setDefaultFont(option.font)
        doc.setDocumentMargin(0)

        text_width = min(self.TEXT_FRACTIONAL_WIDTH * option.rect.width(), doc.idealWidth())
        doc.setTextWidth(text_width)

        return doc

    def paint(
        self, painter: QPainter | None, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        if painter is None:
            return

        style = option.widget.style() if option.widget else QApplication.style()

        if style is None:
            raise ValueError('style is None!')

        doc = self._create_text_document(option, index)
        metrics = BubbleMetrics.from_document(doc)
        alignment = Qt.Alignment(index.data(Qt.ItemDataRole.TextAlignmentRole))

        doc_size = doc.size()
        item_size = QSizeF(
            doc_size.width() + 2 * metrics.mbp_px,
            doc_size.height() + 2 * metrics.mbp_px,
        )
        layout_rect = QStyle.alignedRect(
            Qt.LayoutDirection.LayoutDirectionAuto,
            alignment,
            item_size.toSize(),
            style.subElementRect(QStyle.SubElement.SE_ItemViewItemText, option),
        )
        bubble_rect = QRectF(
            layout_rect.left() + metrics.margin_px,
            layout_rect.top() + metrics.margin_px,
            doc_size.width() + 2 * metrics.bp_px,
            doc_size.height() + 2 * metrics.bp_px,
        )
        text_origin = QPointF(
            bubble_rect.left() + metrics.bp_px,
            bubble_rect.top() + metrics.bp_px,
        )
        text_rect = QRectF(
            0.0,
            0.0,
            doc_size.width(),
            doc_size.height(),
        )

        bubble_brush = QBrush(index.data(Qt.ItemDataRole.BackgroundRole))
        bubble_pen = QPen(index.data(Qt.ItemDataRole.ForegroundRole))
        bubble_pen.setWidth(metrics.border_px)

        style.drawControl(QStyle.ControlElement.CE_ItemViewItem, option, painter, option.widget)

        painter.save()
        painter.setPen(bubble_pen)
        painter.setBrush(bubble_brush)
        painter.drawRoundedRect(bubble_rect, metrics.radius_px, metrics.radius_px)
        painter.translate(text_origin)
        doc.drawContents(painter, text_rect)
        painter.restore()

    def sizeHint(self, option: QStyleOptionViewItem, index: QModelIndex) -> QSize:  # noqa: N802
        doc = self._create_text_document(option, index)
        metrics = BubbleMetrics.from_document(doc)
        hint = option.rect.size()
        hint.setHeight(int(doc.size().height()) + 2 * metrics.mbp_px)
        return hint



from __future__ import annotations

from PyQt5.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMenu,
    QPlainTextEdit,
    QPushButton,
    QTableView,
    QVBoxLayout,
    QWidget,
)


class ProductEditorPropertiesView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Properties')
        self.table_view = QTableView()

        layout = QVBoxLayout()
        layout.addWidget(self.table_view)
        self.setLayout(layout)


class ProductEditorActionsView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Actions')
        self.estimate_probe_photon_count_button = QPushButton('Estimate Probe Photon Count')

        layout = QVBoxLayout()
        layout.addWidget(self.estimate_probe_photon_count_button)
        layout.addStretch()
        self.setLayout(layout)


class ProductEditorCommentsView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Comments')
        self.text_edit = QPlainTextEdit()

        layout = QVBoxLayout()
        layout.addWidget(self.text_edit)
        self.setLayout(layout)


class ProductEditorDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.properties_view = ProductEditorPropertiesView()
        self.actions_view = ProductEditorActionsView()
        self.comments_view = ProductEditorCommentsView()
        self.button_box = QDialogButtonBox()

        self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.button_box.accepted.connect(self.accept)

        top_layout = QHBoxLayout()
        top_layout.addWidget(self.properties_view)
        top_layout.addWidget(self.actions_view)

        layout = QVBoxLayout()
        layout.addLayout(top_layout)
        layout.addWidget(self.comments_view)
        layout.addWidget(self.button_box)
        self.setLayout(layout)

    @property
    def table_view(self) -> QTableView:
        return self.properties_view.table_view

    @property
    def text_edit(self) -> QPlainTextEdit:
        return self.comments_view.text_edit


class ProductButtonBox(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.insert_menu = QMenu()
        self.insert_button = QPushButton('Insert')
        self.save_menu = QMenu()
        self.save_button = QPushButton('Save')
        self.edit_button = QPushButton('Edit')
        self.remove_button = QPushButton('Remove')

        self.insert_button.setMenu(self.insert_menu)
        self.save_button.setMenu(self.save_menu)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.insert_button)
        layout.addWidget(self.save_button)
        layout.addWidget(self.edit_button)
        layout.addWidget(self.remove_button)
        self.setLayout(layout)


class ProductView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.table_view = QTableView()
        self.info_label = QLabel()
        self.button_box = ProductButtonBox()

        layout = QVBoxLayout()
        layout.addWidget(self.table_view)
        layout.addWidget(self.info_label)
        layout.addWidget(self.button_box)
        self.setLayout(layout)



from collections.abc import Sequence
from enum import IntEnum
from typing import overload

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import IntegerParameter


class PtyChiAffineDegreesOfFreedom(IntEnum):
    TRANSLATION = 0
    ROTATION = 1
    SCALING = 2
    SHEARING = 3
    ASYMMETRY = 4


class PtyChiAffineDegreesOfFreedomBitField(Sequence[str], Observable, Observer):
    def __init__(self, parameter: IntegerParameter) -> None:
        super().__init__()
        self._parameter = parameter
        parameter.add_observer(self)

    def is_bit_set(self, bit: int) -> bool:
        value = self._parameter.get_value()
        mask = 1 << bit
        return value & mask != 0

    def set_bit(self, bit: int, is_set: bool) -> None:
        value = self._parameter.get_value()
        mask = 1 << bit

        if is_set:
            value |= mask
        else:
            value &= ~mask

        self._parameter.set_value(value)

    @overload
    def __getitem__(self, index: int) -> str: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[str]: ...

    def __getitem__(self, index: int | slice) -> str | Sequence[str]:
        if isinstance(index, slice):
            return [self[idx] for idx in range(index.start, index.stop, index.step)]

        return PtyChiAffineDegreesOfFreedom(index).name.title()

    def __len__(self) -> int:
        return len(PtyChiAffineDegreesOfFreedom)

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.notify_observers()



"""Wavefield propagation models and associated parameter containers."""

from abc import ABC, abstractmethod
from dataclasses import dataclass

from scipy.fft import fft2, fftfreq, fftshift, ifft2, ifftshift
import numpy

from .common import ComplexArrayType, RealArrayType


def intensity(wavefield: ComplexArrayType) -> RealArrayType:
    return numpy.real(numpy.multiply(wavefield, numpy.conjugate(wavefield)))


@dataclass(frozen=True)
class PropagatorParameters:
    wavelength_m: float
    """Illumination wavelength in meters."""
    width_px: int
    """Number of pixels in the x-direction."""
    height_px: int
    """Number of pixels in the y-direction."""
    pixel_width_m: float
    """Source-plane pixel width in meters."""
    pixel_height_m: float
    """Source-plane pixel height in meters."""
    propagation_distance_m: float
    """Propagation distance in meters."""

    @property
    def dx(self) -> float:
        """Pixel width in wavelengths."""
        return self.pixel_width_m / self.wavelength_m

    @property
    def pixel_aspect_ratio(self) -> float:
        """Pixel aspect ratio (width / height)."""
        return self.pixel_width_m / self.pixel_height_m

    @property
    def z(self) -> float:
        """Propagation distance in wavelengths."""
        return self.propagation_distance_m / self.wavelength_m

    @property
    def fresnel_number(self) -> float:
        """Fresnel number."""
        return numpy.square(self.dx) / numpy.absolute(self.z)

    def get_spatial_coordinates(self) -> tuple[RealArrayType, RealArrayType]:
        JJ, II = numpy.mgrid[: self.height_px, : self.width_px]  # noqa: N806
        XX = II - self.width_px // 2  # noqa: N806
        YY = JJ - self.height_px // 2  # noqa: N806
        return YY, XX

    def get_frequency_coordinates(self) -> tuple[RealArrayType, RealArrayType]:
        fx = fftshift(fftfreq(self.width_px))
        fy = fftshift(fftfreq(self.height_px))
        FY, FX = numpy.meshgrid(fy, fx, indexing='ij')  # noqa: N806
        return FY, FX


class Propagator(ABC):
    """Abstract interface for free-space wavefield propagators."""

    @abstractmethod
    def propagate(self, wavefield: ComplexArrayType) -> ComplexArrayType:
        pass


class AngularSpectrumPropagator(Propagator):
    """Exact propagator using the angular-spectrum transfer function; valid for all Fresnel numbers."""

    def __init__(self, parameters: PropagatorParameters) -> None:
        ar = parameters.pixel_aspect_ratio

        i2piz = 2j * numpy.pi * parameters.z
        FY, FX = parameters.get_frequency_coordinates()  # noqa: N806
        F2 = numpy.square(FX) + numpy.square(ar * FY)  # noqa: N806
        ratio = F2 / numpy.square(parameters.dx)
        tf = numpy.exp(i2piz * numpy.sqrt(1 - ratio))

        self._transfer_function = numpy.where(ratio < 1, tf, 0)

    def propagate(self, wavefield: ComplexArrayType) -> ComplexArrayType:
        return fftshift(ifft2(self._transfer_function * fft2(ifftshift(wavefield))))


class FresnelTransferFunctionPropagator(Propagator):
    """Fresnel propagator using a paraxial transfer function in the frequency domain."""

    def __init__(self, parameters: PropagatorParameters) -> None:
        ar = parameters.pixel_aspect_ratio

        i2piz = 2j * numpy.pi * parameters.z
        FY, FX = parameters.get_frequency_coordinates()  # noqa: N806
        F2 = numpy.square(FX) + numpy.square(ar * FY)  # noqa: N806
        ratio = F2 / numpy.square(parameters.dx)

        self._transfer_function = numpy.exp(i2piz * (1 - ratio / 2))

    def propagate(self, wavefield: ComplexArrayType) -> ComplexArrayType:
        return fftshift(ifft2(self._transfer_function * fft2(ifftshift(wavefield))))


class FresnelTransformPropagator(Propagator):
    """Fresnel propagator using the direct Fresnel transform; changes pixel size between planes."""

    def __init__(self, parameters: PropagatorParameters) -> None:
        ipi = 1j * numpy.pi

        Fr = parameters.fresnel_number  # noqa: N806
        ar = parameters.pixel_aspect_ratio
        N = parameters.width_px  # noqa: N806
        M = parameters.height_px  # noqa: N806
        YY, XX = parameters.get_spatial_coordinates()  # noqa: N806

        C0 = Fr / (1j * ar)  # noqa: N806
        C1 = numpy.exp(2j * numpy.pi * parameters.z)  # noqa: N806
        C2 = numpy.exp((numpy.square(XX / N) + numpy.square(ar * YY / M)) * ipi / Fr)  # noqa: N806
        is_forward = parameters.propagation_distance_m >= 0.0

        self._is_forward = is_forward
        self._A = C2 * C1 * C0 if is_forward else C2 * C1 / C0
        self._B = numpy.exp(ipi * Fr * (numpy.square(XX) + numpy.square(YY / ar)))

    def propagate(self, wavefield: ComplexArrayType) -> ComplexArrayType:
        if self._is_forward:
            return self._A * fftshift(fft2(ifftshift(wavefield * self._B)))
        else:
            return self._B * fftshift(ifft2(ifftshift(wavefield * self._A)))


class FraunhoferPropagator(Propagator):
    """Far-field (Fraunhofer) propagator; valid when the Fresnel number is much less than one."""

    def __init__(self, parameters: PropagatorParameters) -> None:
        ipi = 1j * numpy.pi

        Fr = parameters.fresnel_number  # noqa: N806
        ar = parameters.pixel_aspect_ratio
        N = parameters.width_px  # noqa: N806
        M = parameters.height_px  # noqa: N806
        YY, XX = parameters.get_spatial_coordinates()  # noqa: N806

        C0 = Fr / (1j * ar)  # noqa: N806
        C1 = numpy.exp(2j * numpy.pi * parameters.z)  # noqa: N806
        C2 = numpy.exp((numpy.square(XX / N) + numpy.square(ar * YY / M)) * ipi / Fr)  # noqa: N806
        is_forward = parameters.propagation_distance_m >= 0.0

        self._is_forward = is_forward
        self._A = C2 * C1 * C0 if is_forward else C2 * C1 / C0

    def propagate(self, wavefield: ComplexArrayType) -> ComplexArrayType:
        if self._is_forward:
            return self._A * fftshift(fft2(ifftshift(wavefield)))
        else:
            return fftshift(ifft2(ifftshift(wavefield * self._A)))



from PyQt5.QtCore import QEvent, QModelIndex, QObject, Qt
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QFormLayout,
    QGroupBox,
    QInputDialog,
    QListView,
    QPushButton,
    QVBoxLayout,
)

from ...model.agent import (
    AgentPresenter,
    ArgoSettings,
    ChatHistory,
    ChatMessage,
    ChatObserver,
)
from ...view.agent import AgentChatView, AgentInputView, AgentView
from ..parametric import (
    ComboBoxParameterViewController,
    DecimalSliderParameterViewController,
    LineEditParameterViewController,
    SpinBoxParameterViewController,
)
from .item_delegate import ChatBubbleItemDelegate
from .list_model import AgentMessageListModel

__all__ = ['AgentChatController', 'AgentController']


class AgentInputController(QObject):
    def __init__(self, presenter: AgentPresenter, view: AgentInputView) -> None:
        super().__init__()
        self._presenter = presenter
        self._view = view

        view.text_edit.installEventFilter(self)
        view.send_button.clicked.connect(self._send_message)

    def _send_message(self) -> None:
        text = self._view.text_edit.toPlainText()
        self._presenter.send_message(text)
        self._view.text_edit.clear()

    def eventFilter(self, a0: QObject, a1: QEvent) -> bool:  # noqa: N802
        if a0 == self._view.text_edit and isinstance(a1, QKeyEvent):
            is_shift_pressed = bool(a1.modifiers() & Qt.KeyboardModifier.ShiftModifier)

            # require shift+enter for new line, otherwise send on enter
            if a1.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return) and not is_shift_pressed:
                self._send_message()
                return True

        return super().eventFilter(a0, a1)


class AgentChatController(ChatObserver):
    def __init__(
        self, history: ChatHistory, presenter: AgentPresenter, view: AgentChatView
    ) -> None:
        super().__init__()
        self._history = history
        self._presenter = presenter
        self._view = view
        self._message_list_model = AgentMessageListModel(history)
        self._input_controller = AgentInputController(presenter, view.input_view)

        view.message_list_view.setModel(self._message_list_model)
        view.message_list_view.setItemDelegate(ChatBubbleItemDelegate())
        view.message_list_view.setResizeMode(QListView.ResizeMode.Adjust)
        view.message_list_view.setVerticalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)

        history.add_observer(self)

    def handle_new_message(self, message: ChatMessage, index: int) -> None:
        parent = QModelIndex()
        self._message_list_model.beginInsertRows(parent, index, index)
        self._message_list_model.endInsertRows()

    def handle_chat_cleared(self) -> None:
        self._message_list_model.beginResetModel()
        self._message_list_model.endResetModel()


class AgentController:
    def __init__(self, settings: ArgoSettings, presenter: AgentPresenter, view: AgentView) -> None:
        self._settings = settings
        self._presenter = presenter
        self._view = view

        self._user_view_controller = LineEditParameterViewController(settings.user)
        self._chat_endpoint_url_view_controller = LineEditParameterViewController(
            settings.chat_endpoint_url, tool_tip='The chat endpoint URL.'
        )
        self._chat_model_view_controller = ComboBoxParameterViewController(
            settings.chat_model,
            presenter.get_available_chat_models(),
            tool_tip='The chat model to use.',
        )
        self._temperature_view_controller = DecimalSliderParameterViewController(
            settings.temperature,
            tool_tip='What sampling temperature to use, between 0 and 2. Higher values mean the model takes more risks.',
        )
        self._top_p_view_controller = DecimalSliderParameterViewController(
            settings.top_p,
            tool_tip='An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass.',
        )
        self._max_tokens_view_controller = SpinBoxParameterViewController(
            settings.max_tokens,
            tool_tip='The maximum number of tokens that can be generated in the chat completion.',
        )
        self._max_completion_tokens_view_controller = SpinBoxParameterViewController(
            settings.max_completion_tokens,
            tool_tip='An upper bound for the number of tokens that can be generated for a completion, including visible output tokens and reasoning tokens.',
        )
        self._embeddings_endpoint_url_view_controller = LineEditParameterViewController(
            settings.embeddings_endpoint_url, tool_tip='The embeddings endpoint URL.'
        )
        self._embeddings_model_view_controller = ComboBoxParameterViewController(
            settings.embeddings_model,
            presenter.get_available_embeddings_models(),
            tool_tip='The embeddings model to use.',
        )
        self._embed_button = QPushButton('Embed Text')
        self._embed_button.clicked.connect(self._embed_text)

        group_box_layout = QFormLayout()
        group_box_layout.addRow('User:', self._user_view_controller.get_widget())
        group_box_layout.addRow(
            'Chat Endpoint URL:', self._chat_endpoint_url_view_controller.get_widget()
        )
        group_box_layout.addRow('Chat Model:', self._chat_model_view_controller.get_widget())
        group_box_layout.addRow('Temperature:', self._temperature_view_controller.get_widget())
        group_box_layout.addRow('Top P:', self._top_p_view_controller.get_widget())
        group_box_layout.addRow('Max Tokens:', self._max_tokens_view_controller.get_widget())
        group_box_layout.addRow(
            'Max Completion Tokens:', self._max_completion_tokens_view_controller.get_widget()
        )
        group_box_layout.addRow(
            'Embeddings Endpoint URL:', self._embeddings_endpoint_url_view_controller.get_widget()
        )
        group_box_layout.addRow(
            'Embeddings Model:', self._embeddings_model_view_controller.get_widget()
        )
        group_box_layout.addRow(self._embed_button)

        group_box = QGroupBox('Argo')
        group_box.setLayout(group_box_layout)

        layout = QVBoxLayout()
        layout.addWidget(group_box)
        layout.addStretch()
        view.setLayout(layout)

    def _embed_text(self) -> None:
        title = 'Embed Text'
        label = 'Enter text to embed:'
        text, ok_pressed = QInputDialog.getMultiLineText(self._view, title, label, text='')

        if ok_pressed:
            self._presenter.embed_text(text.splitlines())



from __future__ import annotations

from PyQt5.QtWidgets import QAction, QActionGroup, QStatusBar

from ptychodus.api.common import NumberArrayType
from ptychodus.api.geometry import PixelGeometry

from ...model.visualization import VisualizationEngine
from ...view.visualization import ImageItem, ImageMouseTool, VisualizationWidget
from ..data import FileDialogFactory
from .controller import VisualizationController


class VisualizationWidgetController:
    def __init__(
        self,
        engine: VisualizationEngine,
        widget: VisualizationWidget,
        status_bar: QStatusBar,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        self._widget = widget
        self._controller = VisualizationController(
            engine, widget.visualization_view, status_bar, file_dialog_factory
        )

        widget.home_action.triggered.connect(self._controller.zoom_to_fit)
        widget.save_action.triggered.connect(self._controller.save_image)
        widget.autoscale_action.triggered.connect(self._auto_display_range)

        widget.move_action.setData(ImageMouseTool.MOVE_TOOL.value)
        widget.move_action.setCheckable(True)
        widget.move_action.setChecked(True)

        widget.ruler_action.setData(ImageMouseTool.RULER_TOOL.value)
        widget.ruler_action.setCheckable(True)

        widget.rectangle_action.setData(ImageMouseTool.RECTANGLE_TOOL.value)
        widget.rectangle_action.setCheckable(True)

        widget.line_cut_action.setData(ImageMouseTool.LINE_CUT_TOOL.value)
        widget.line_cut_action.setCheckable(True)

        self._mouse_tool_action_group = QActionGroup(widget.tool_bar)
        self._mouse_tool_action_group.addAction(widget.move_action)
        self._mouse_tool_action_group.addAction(widget.ruler_action)
        self._mouse_tool_action_group.addAction(widget.rectangle_action)
        self._mouse_tool_action_group.addAction(widget.line_cut_action)
        self._mouse_tool_action_group.triggered.connect(self._set_mouse_tool)

    def get_item(self) -> ImageItem:
        return self._controller.get_item()

    def _auto_display_range(self) -> None:
        self._controller.rerender_image(autoscale_color_axis=True)

    def set_array(self, array: NumberArrayType, pixel_geometry: PixelGeometry) -> None:
        self._controller.set_array(array, pixel_geometry)

    def clear_array(self) -> None:
        self._controller.clear_array()

    def _set_mouse_tool(self, mouse_tool_action: QAction) -> None:
        tool_id = mouse_tool_action.data()
        mouse_tool = ImageMouseTool(tool_id)
        self._controller.set_mouse_tool(mouse_tool)



from .average_pattern import AveragePatternProbeBuilder
from .builder import ProbeSequenceBuilder
from .builder_factory import ProbeBuilderFactory
from .disk import DiskProbeBuilder
from .fzp import FresnelZonePlateProbeBuilder
from .item import ProbeRepositoryItem
from .item_factory import ProbeRepositoryItemFactory
from .multimodal import MultimodalProbeBuilder, ProbeModeDecayType
from .rect import RectangularProbeBuilder
from .settings import ProbeSettings
from .super_gaussian import SuperGaussianProbeBuilder
from .zernike import ZernikeProbeBuilder

__all__ = [
    'AveragePatternProbeBuilder',
    'DiskProbeBuilder',
    'FresnelZonePlateProbeBuilder',
    'MultimodalProbeBuilder',
    'ProbeSequenceBuilder',
    'ProbeBuilderFactory',
    'ProbeModeDecayType',
    'ProbeRepositoryItem',
    'ProbeRepositoryItemFactory',
    'ProbeSettings',
    'RectangularProbeBuilder',
    'SuperGaussianProbeBuilder',
    'ZernikeProbeBuilder',
]



from __future__ import annotations
from collections.abc import Iterator

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.plugins import Plugin, PluginChooser
from ptychodus.api.settings import SettingsRegistry
from ptychodus.api.workflow import FileBasedWorkflow, WorkflowAPI

from ..task_manager import TaskManager
from .buffer import PendingFileBuffer
from .repository import AutomationDatasetRepository
from .settings import AutomationSettings
from .watcher import DataDirectoryWatcher


class AutomationPresenter(Observable, Observer):
    def __init__(
        self,
        settings: AutomationSettings,
        watcher: DataDirectoryWatcher,
        dataset_buffer: PendingFileBuffer,
        workflow_chooser: PluginChooser[FileBasedWorkflow],
    ) -> None:
        super().__init__()
        self._settings = settings
        self._watcher = watcher
        self._dataset_buffer = dataset_buffer
        self._workflow_chooser = workflow_chooser

        watcher.add_observer(self)

    @property
    def _current_workflow_plugin(self) -> Plugin[FileBasedWorkflow]:
        return self._workflow_chooser.get_current_plugin()

    @property
    def _current_workflow(self) -> FileBasedWorkflow:
        return self._current_workflow_plugin.strategy

    def available_workflows(self) -> Iterator[str]:
        for plugin in self._workflow_chooser:
            yield plugin.display_name

    def get_current_workflow(self) -> str:
        return self._current_workflow_plugin.display_name

    def set_current_workflow(self, name: str) -> None:
        self._workflow_chooser.set_current_plugin(name)

    def load_existing_datasets_to_repository(self) -> None:
        data_directory = self._settings.data_directory.get_value()
        pattern = '**/' if self._current_workflow_plugin.strategy.is_watch_recursive else ''
        pattern += self._current_workflow_plugin.strategy.get_watch_file_pattern()
        scan_file_list = sorted(scanFile for scanFile in data_directory.glob(pattern))

        for scan_file in scan_file_list:
            self._dataset_buffer.put(scan_file)

    def is_watcher_enabled(self) -> bool:
        return self._watcher.is_alive

    def set_watcher_enabled(self, enable: bool) -> None:
        if enable:
            self._watcher.start()
        else:
            self._watcher.stop()

    def _update(self, observable: Observable) -> None:
        if observable is self._watcher:
            self.notify_observers()


class AutomationCore:
    def __init__(
        self,
        task_manager: TaskManager,
        settings_registry: SettingsRegistry,
        workflow_api: WorkflowAPI,
        workflow_chooser: PluginChooser[FileBasedWorkflow],
    ) -> None:
        self.settings = AutomationSettings(settings_registry)
        self.repository = AutomationDatasetRepository(self.settings)
        self._file_buffer = PendingFileBuffer(
            task_manager, self.settings, self.repository, workflow_api, workflow_chooser
        )
        self._watcher = DataDirectoryWatcher(self.settings, self._file_buffer)
        self.presenter = AutomationPresenter(
            self.settings, self._watcher, self._file_buffer, workflow_chooser
        )

    def start(self) -> None:
        self._file_buffer.start()

    def stop(self) -> None:
        self._watcher.stop()
        self._file_buffer.stop()



import logging

from ptychodus.api.settings import SettingsRegistry

from ..diffraction import AssembledDiffractionDataset
from ..product import ObjectRepository, ProductRepository
from ..visualization import VisualizationEngine
from .diffraction import DiffractionSimulator
from .fourier import FourierAnalyzer
from .frc import FourierRingCorrelator
from .illumination import IlluminationMapper
from .propagator import ProbePropagator
from .settings import ProbePropagationSettings
from .xmcd import XMCDAnalyzer

logger = logging.getLogger(__name__)


class AnalysisCore:
    def __init__(
        self,
        settings_registry: SettingsRegistry,
        dataset: AssembledDiffractionDataset,
        product_repository: ProductRepository,
        object_repository: ObjectRepository,
    ) -> None:
        self.diffraction_simulator = DiffractionSimulator(dataset, product_repository)
        self._probe_propagation_settings = ProbePropagationSettings(settings_registry)
        self.probe_propagator = ProbePropagator(
            self._probe_propagation_settings, product_repository
        )
        self.probe_propagator_visualization_engine = VisualizationEngine(is_complex=False)

        self.exposure_analyzer = IlluminationMapper(product_repository)
        self.exposure_visualization_engine = VisualizationEngine(is_complex=False)

        self.fourier_ring_correlator = FourierRingCorrelator(object_repository)

        self.fourier_analyzer = FourierAnalyzer(product_repository)
        self.fourier_real_space_visualization_engine = VisualizationEngine(is_complex=True)
        self.fourier_reciprocal_space_visualization_engine = VisualizationEngine(is_complex=True)

        self.xmcd_analyzer = XMCDAnalyzer(product_repository)
        self.xmcd_visualization_engine = VisualizationEngine(is_complex=False)



from .core import (
    PtychoNNModelSettings,
    PtychoNNReconstructorLibrary,
    PtychoNNTrainingSettings,
)

__all__ = [
    'PtychoNNModelSettings',
    'PtychoNNReconstructorLibrary',
    'PtychoNNTrainingSettings',
]
